from datahub import *
import pandas as pd
backend = "sf-databuffer"
channels = ["S10BC01-DBPM010:Q1", "S10BC01-DBPM010:X1"]
start = "2024-06-14 09:00:00"
end = "2024-06-14 10:00:00"
bins = None
df = None
query = None

def fetch_data(_bins):
    global df, bins, query
    bins = _bins
    query = {
        "channels": channels,
        "start": start,
        "end": end,
    }
    if (bins):
        query["bins"] = bins

    with Daqbuf(backend=backend, cbor=True, parallel=True, time_type="seconds") as source:
        table = Table()
        source.add_listener(table)
        source.request(query)
        #dataframe_cbor = table.as_dataframe(Table.PULSE_ID)
        df = table.as_dataframe(Table.TIMESTAMP)
        df.reindex(sorted(df.columns), axis=1)

fetch_data(100)

import dash
import dash_core_components as dcc
import dash_html_components as html
from dash import Dash, html, dcc, callback, Output, Input, State
import plotly.graph_objects as go
import numpy as np


def get_figure(df, channel, color):
    cols = list(df.columns)
    binned = (channel + " max") in cols
    x= pd.to_datetime(df.index, unit='s')
    y=df[channel]
    return {
        'data': [
            go.Scatter(
                x=x,
                y=y,
                mode='lines+markers',
                name=channel,
                line=dict(color=f'rgba({color[0]}, {color[1]}, {color[2]}, 1.0)'),
                marker=dict(
                    color=f'rgba({color[0]}, {color[1]}, {color[2]}, 1.0)',
                    size=5
                ),
                error_y=None if not binned else dict(
                    type='data',
                    symmetric=False,
                    array= df[channel + " max"] - y,
                    arrayminus= y - df[channel + " min"] ,
                    #array= [1.0] * len(df.index),
                    # arrayminus= [2.0] * len(df.index),
                    visible=True,
                    color=f'rgba({color[0]}, {color[1]}, {color[2]}, 0.1)'  # Set the color with transparency
                )
            )
        ],
        'layout': go.Layout(
            title=str(query),
            xaxis={'title': None},
            yaxis={'title': channel}
        )
    }

def fetch_graphs():
    return [
        dcc.Dropdown(channels, channels[0], id='dropdown-selection'),
        dcc.Graph(
            id='channel_graph',
            #figure= get_figure(df, channels[0], c)
        )
    ]


# Create the Dash app
app = dash.Dash(__name__, suppress_callback_exceptions=True)
c = (0 ,0, 255)
# Define the layout of the app
app.layout = html.Div(children=[
    html.H1(children='Daqbuf UI'),
    html.Div([
        html.Label('Bins:', style={'margin-right': '10px'}),
        dcc.Input(id='input_bins', type='number', min=0, max=1000, step=1, style={'margin-right': '10px'}, value=bins),
        html.Button('Submit', id='button')
    ], style={'display': 'flex', 'align-items': 'center', 'margin-bottom': '10px'}),
    html.Div(id='output-container', style={'margin-top': '20px'})
])




# Define the callback
@callback(
    Output('output-container', 'children'),
    Input('button', 'n_clicks'),
    State('input_bins', 'value')
)
def update_output(n_clicks, bins):
    if n_clicks is None:
        return ''
    fetch_data(bins)
    return fetch_graphs()



@callback(
    Output('channel_graph', 'figure'),
    Input('dropdown-selection', 'value')
)
def update_graph(value):
    return get_figure(df, value, c)

# Run the app
if __name__ == '__main__':
    app.run_server(debug=True)



from datahub import *
import io
from threading import Thread
from http.client import IncompleteRead


_logger = logging.getLogger(__name__)

class Daqbuf(Source):
    """
    Retrieves data from a Daqbuf service (new retrieval).
    """

    DEFAULT_URL = os.environ.get("DAQBUF_DEFAULT_URL", "https://data-api.psi.ch/api/4")
    DEFAULT_BACKEND = os.environ.get("DAQBUF_DEFAULT_BACKEND", "sf-databuffer")

    def __init__(self, url=DEFAULT_URL, backend=DEFAULT_BACKEND, path=None, delay=1.0, cbor=True, parallel=False, **kwargs):
        """
        url (str, optional): Daqbuf URL. Default value can be set by the env var DAQBUF_DEFAULT_URL.
        backend (str, optional): Daqbuf backend. Default value can be set by the env var DAQBUF_DEFAULT_BACKEND.
        path (str, optional): hint for the source location in storage or displaying.
        delay (float, optional): Wait time for channels to be uploaded to storage before retrieval.
        cbor (bool, optional): if True (default) retrieves data as CBOR, otherwise as JSON.
        parallel (bool, optional): if True performs the retrieval of multiple channels in differt threads.
        """
        if url is None:
            raise RuntimeError("Invalid URL")
        Source.__init__(self, url=url, backend=backend, query_path="/events",  search_path="/search/channel", path=path,
                        known_backends=None, **kwargs)
        self.base_url = url
        self.binned_url = self.base_url + "/binned"
        self.known_backends = self.get_backends()
        self.delay = delay
        self.cbor = str_to_bool(str(cbor))
        self.parallel = str_to_bool(str(parallel))
        if self.cbor:
            try:
                import cbor2
                self.cbor = cbor2
            except:
                _logger.error("cbor2 not installed: JSON fallback on Daqbuf searches")
                self.cbor = None

    def pulse_id_to_time(self, id):
        import requests
        response = requests.get(self.base_url + "/map/pulse/sf-databuffer/" + str(id))
        if response.status_code != 200:
            raise RuntimeError("Unable to retrieve data from server: ", response)
        data = response.text
        nanos = int(data)
        secs = convert_timestamp(nanos, "sec")
        ret = round(secs, PULSE_ID_INTERVAL_DEC)
        return ret

    def get_backends(self):
        try:
            if self.known_backends is None:
                import requests
                response = requests.get(self.base_url + "/backend/list")
                ret = response.json()
                backends = ret["backends_available"]
                self.known_backends = [backend["name"] for backend in backends]
            return self.known_backends
        except Exception as e:
            _logger.exception(e)
            return []

    def read(self, stream, channel):
        try:
            current_channel_name = None

            while True:
                bytes_read = stream.read(4)
                if len(bytes_read) != 4:
                    break
                length = struct.unpack('<i', bytes_read)[0]

                bytes_read = stream.read(12) #PADDING
                if len(bytes_read) != 12:
                    break

                bytes_read = stream.read(length)
                if len(bytes_read) != length:
                    raise RuntimeError("unexpected EOF")
                parsed_data = self.cbor.loads(bytes_read)

                padding = padding = (8 - (length % 8)) % 8
                bytes_read = stream.read(padding) #PADDING
                if len(bytes_read) != padding:
                    break

                if type(parsed_data) != dict:
                    raise RuntimeError("Invalid cbor frame: " + str(type(parsed_data)))

                if parsed_data.get("error", None) :
                    raise Exception(parsed_data.get("error"))

                if not parsed_data.get ("type","") == 'keepalive':
                    tss = parsed_data.get('tss', [])
                    pulses = parsed_data.get('pulses', [])
                    values = parsed_data.get('values', [])
                    scalar_type = parsed_data.get('scalar_type', None)
                    rangeFinal = parsed_data.get('rangeFinal', False)

                    if scalar_type:
                        nelm = len(values)
                        for i in range(nelm):
                            timestamp = tss[i]
                            pulse_id = pulses[i]
                            value = values[i]
                            self.receive_channel(channel, value, timestamp, pulse_id, check_changes=False, check_types=True)
                            current_channel_name = channel
                    if rangeFinal:
                        break
                    elif not scalar_type:
                        raise RuntimeError("Invalid cbor frame keys: " + str(parsed_data.keys()))

                    if not self.is_running() or self.is_aborted():
                        raise RuntimeError("Query has been aborted")

        except IncompleteRead:
            _logger.error("Unexpected end of input")
            raise ProtocolError()
        finally:
            if current_channel_name:
                self.on_channel_completed(current_channel_name)

    def run_channel(self, channel, cbor, bins=None, last=None, conn=None):
        query = dict()
        query["channelName"] = channel
        query["begDate"] = self.range.get_start_str_iso()
        query["endDate"] = self.range.get_end_str_iso()
        query["backend"] = self.backend
        if last is not None:
            query["oneBeforeRange"] = "true" if last else "false"

        if cbor:
            create_connection = conn is None
            conn = http_data_query(query, self.url, method="GET", accept="application/cbor-framed", conn=conn)
            try:
                response = conn.getresponse()
                if response.status != 200:
                    raise RuntimeError(f"Unable to retrieve data from server: {response.reason} [{response.status}]")
                try:
                    self.read(io.BufferedReader(response), channel)
                except Exception as e:
                    _logger.exception(e)
                    raise
            finally:
                if create_connection:
                    conn.close()

        else:
            import requests
            if bins:
                query["binCount"] = bins
                response = requests.get(self.binned_url, query)
                # Check for successful return of data
                if response.status_code != 200:
                    raise RuntimeError("Unable to retrieve data from server: ", response)
                data = response.json()
                nelm = len(data['avgs'])
                for i in range(nelm):
                    secs1 = data['tsAnchor'] + float(data['ts1Ms'][i]) / 1000.0
                    timestamp1 = create_timestamp(secs1, data['ts1Ns'][i])
                    secs2 = data['tsAnchor'] + float(data['ts2Ms'][i]) / 1000.0
                    timestamp2 = create_timestamp(secs2, data['ts2Ns'][i])
                    avg = data['avgs'][i]
                    max = self.adjust_type(data['maxs'][i])
                    min = self.adjust_type(data['mins'][i])
                    count = self.adjust_type(data['counts'][i])
                    start = self.convert_time(timestamp1)
                    end = self.convert_time(timestamp2)

                    value = avg
                    timestamp = int((timestamp1 + timestamp2)/2)
                    args = {"bins":bins, "min":numpy.float64(min), "max":numpy.float64(max), "count":numpy.int64(count), "start":start, "end": end}
                    self.receive_channel(channel, value, timestamp, None, check_changes=False, check_types=True, metadata={"bins":bins}, **args)
            else:
                response = requests.get(self.url, query)
                # Check for successful return of data
                if response.status_code != 200:
                    raise RuntimeError("Unable to retrieve data from server: ", response)
                data = response.json()
                nelm = len(data['values'])
                for i in range(nelm):
                    secs = data['tsAnchor'] + float(data['tsMs'][i]) / 1000.0
                    timestamp = create_timestamp(secs, data['tsNs'][i])
                    pulse_id = data['pulseAnchor'] + data['pulseOff'][i]
                    value = data['values'][i]
                    self.receive_channel(channel, value, timestamp, pulse_id, check_changes=False, check_types=True)
            self.on_channel_completed(channel)


    def run(self, query):
        self.range.wait_end(delay=self.delay)
        channels = query.get("channels", [])
        bins = query.get("bins", None)
        last = query.get("last", None)
        cbor = self.cbor and not bins
        if isinstance(channels, str):
            channels = [channels, ]
        conn = None
        threads = []
        try:
            if self.parallel:
                for channel in channels:
                    thread = Thread(target=self.run_channel, args=(channel, cbor, bins, last))
                    thread.setDaemon(True)
                    thread.start()
                    threads.append(thread)
                for thread in threads:
                    thread.join()
            else:
                if cbor:
                    conn = create_http_conn(self.url)
                for channel in channels:
                    self.run_channel(channel, cbor, bins, last, conn)
        finally:
            if conn:
                conn.close()
            self.close_channels()

    def search(self, regex):
        import requests
        if not regex:
            return self.get_backends()
        else:
            cfg = {
                "nameRegex": regex
            }
            if self.backend:
                cfg["backend"] = self.backend
            response = requests.get(self.search_url, params=cfg)
            ret = response.json()

            if not self.verbose:
                channels = ret.get("channels", [])
                pd = self._get_pandas()
                if pd is None:
                    ret = [d["name"] for d in ret.get("channels", [])]
                else:
                    if (len(channels)>0):
                        header = list(channels[0].keys()) if len(channels) > 0 else []
                        data = [d.values() for d in channels]
                        df = pd.DataFrame(data, columns=header)
                        df = df.sort_values(by=["backend", "name"])
                        columns_to_display = ["backend", "name", "seriesId", "type", "shape"]
                        ret = df[columns_to_display].to_string(index=False)
                    else:
                        return None
            return ret




import secrets
import string
import collections
import pickle

class MaxLenDict(collections.OrderedDict):
    def __init__(self, *args, **kwds):
        self.maxlen = kwds.pop("maxlen", None)
        collections.OrderedDict.__init__(self, *args, **kwds)
        self._check_maxlen()

    def __setitem__(self, key, value):
        collections.OrderedDict.__setitem__(self, key, value)
        self._check_maxlen()

    def _check_maxlen(self):
        if self.maxlen is not None:
            while len(self) > self.maxlen:
                self.popitem(last=False)


def generate_random_string(length=16):
    characters = string.ascii_letters + string.digits
    random_string = ''.join(secrets.choice(characters) for _ in range(length))
    return random_string


def decode(data):
    obj = pickle.loads(data)
    return obj

def encode(obj):
    value_serialized = pickle.dumps(obj)
    return value_serialized


import logging
import h5py
import numpy
import datetime
import threading
from datahub import Consumer, Compression, bitshuffle_compression_lz4, decompress

_logger = logging.getLogger(__name__)

class HDF5Writer(Consumer):

    def __init__(self, filename: str, default_compression=Compression.GZIP, auto_decompress=False, path=None, metadata_compression=Compression.GZIP, **kwargs):
        Consumer.__init__(self, **kwargs)
        self.nbytes_read = 0
        self.filename = filename
        self.default_compression = default_compression
        self.auto_decompress = auto_decompress
        self.metadata_compression = metadata_compression
        self.in_channel = False
        self.path = ""
        self.file = None
        if path:
            if not path.startswith("/"):
                path = "/" + path
            self.path = path
        self.datasets = {}
        self.lock = threading.Lock()

    def on_start(self, source):
        with self.lock:
            if self.file is None:
                self.file = h5py.File(self.filename, "w")
                now_date = datetime.datetime.now(datetime.timezone.utc)
                self.file.attrs["creation"] = now_date.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"

    def on_stop(self, source, exception):
        pass

    def get_path(self, source):
        if self.path and not source.path:
            return self.path
        return f"{source.get_path()}"


    def get_group(self, source, channel):
        prefix = self.get_path(source)
        for i in range(1000):
            if i==0:
               name = channel
            else:
               name = f"{channel}_{i}"
            group = f"{prefix}/{name}"
            if not group in self.file:
                break
        return name

    def _get_time_fmt(self, source):
        if source.time_type == "str":
            return "str"
        elif source.time_type == "sec":
            return numpy.float64
        else:
            return numpy.int64

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        prefix, channel = self.get_path(source), self.get_group(source, name)
        has_id = metadata.get("has_id", True)
        ts_ds = Dataset(prefix, channel, "timestamp", self.file, dtype=self._get_time_fmt(source), dataset_compression=self.metadata_compression)
        id_ds = Dataset(prefix, channel, "id", self.file, dtype=numpy.int64, dataset_compression=self.metadata_compression) if has_id else None
        data_ds_name = "value"
        if channel_compression and (not self.auto_decompress):
            if shape is None or (len(shape) == 0):
                raise RuntimeError(f"Compression not supported on scalars")
            if channel_compression != Compression.BITSHUFFLE_LZ4:
                raise RuntimeError(f"Compression not supported: " + channel_compression)
            val_ds = DirectChunkWriteDataset(prefix, channel, data_ds_name, self.file,shape, typ, channel_compression, dataset_compression=Compression.BITSHUFFLE_LZ4)
        else:
            val_ds = Dataset(prefix, channel, data_ds_name, self.file, shape, typ, channel_compression, dataset_compression=self.default_compression)
            if metadata.get("bins", None):
                min_ds = Dataset(prefix,  channel, "min", self.file, shape, typ, channel_compression, dataset_compression=self.default_compression)
                max_ds = Dataset(prefix, channel, "max", self.file, shape, typ, channel_compression,dataset_compression=self.default_compression)
                cnt_ds = Dataset(prefix, channel, "count", self.file, dtype=numpy.int64, dataset_compression=self.metadata_compression)
                start_ds = Dataset(prefix, channel, "start", self.file, dtype=self._get_time_fmt(source),dataset_compression=self.metadata_compression)
                end_ds = Dataset(prefix, channel, "end", self.file, dtype=self._get_time_fmt(source),dataset_compression=self.metadata_compression)
                val_ds = val_ds, min_ds, max_ds, cnt_ds, start_ds, end_ds

        if not self.datasets.get(source, None):
            self.datasets[source] = {}
            self.file[f"{prefix}"].attrs["type"] = str(source.type)
            self.file[f"{prefix}"].attrs["backend"] = str(source.backend)
            self.file[f"{prefix}"].attrs["url"] = str(source.url)
            self.file[f"{prefix}"].attrs["query_index"] = str(source.query_index)
            self.file[f"{prefix}"].attrs["name"] = str(source.get_name())
            for key in source.query.keys():
                self.file[f"{prefix}"].attrs[key] = str(source.query[key])
        self.datasets[source][name] = [ts_ds, id_ds, val_ds]
        self.file[f"{prefix}/{channel}"].attrs["name"] = str(name)
        self.file[f"{prefix}/{channel}"].attrs["type"] = str(type)
        self.file[f"{prefix}/{channel}"].attrs["byteOrder"] = str(byteOrder)
        self.file[f"{prefix}/{channel}"].attrs["shape"] = str(shape)
        self.file[f"{prefix}/{channel}"].attrs["compression"] = str(channel_compression)
        for key in metadata.keys():
            self.file[f"{prefix}/{channel}"].attrs[key] = metadata[key]

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        [ts_ds, id_ds, val_ds] = self.datasets[source][name]
        if ts_ds:
            ts_ds.append(timestamp)
        if id_ds:
            id_ds.append(pulse_id)
        if kwargs.get("bins", None):
            val_ds, min_ds, max_ds, cnt_ds, start_ds, end_ds = val_ds
            min_ds.append(kwargs.get("min"))
            max_ds.append(kwargs.get("max"))
            cnt_ds.append(kwargs.get("count"))
            start_ds.append(kwargs.get("start"))
            end_ds.append(kwargs.get("end"))
        val_ds.append(value)

    def on_channel_completed(self, source, name):
        self.close_datasets(source, name)

    def close_datasets(self, source=None, name=None):
        if source is None:
            for source in list(self.datasets.keys()):
                self.close_datasets(source, None)
            return
        if name is None:
            for name in list(self.datasets[source].keys()):
                self.close_datasets(source, name)
            return
        for dataset in self.datasets[source].get(name, []):
            try:
                if dataset is not None:
                    if type(dataset) is tuple:
                        for d in dataset:
                            d.close()
                    else:
                        dataset.close()
            except Exception as ex:
                _logger.exception("Error closing datasets of channel %s: %s " % (name, str(ex)))

        self.datasets[source].pop(name, None)

    def on_close(self):
        self.close_datasets()
        try:
            if self.file:
                now_date = datetime.datetime.now(datetime.timezone.utc)
                self.file.attrs["conclusion"] = now_date.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
                self.file.close()
        except Exception as ex:
            _logger.exception("Error closing file: %s " % str(self.file_name))
        self.file = None




class Dataset:
    STRING_TYPE = h5py.string_dtype()
    def __init__(self, prefix, channel, field, h5file, shape=None, dtype=None, channel_compression=None, chunks=None, dataset_compression=Compression.GZIP, compression_opts=None, shuffle=True):
        self.channel_compression = channel_compression
        self.dataset_compression = dataset_compression
        self.compression_opts=compression_opts
        self.channel = channel
        self.h5file = h5file
        if shape is None:
            shape = tuple()
        shape = tuple(shape)
        self.shape = shape

        if dtype is None:
            dtype = int
        elif dtype == "str":
            dtype = str
        self.is_value = type(dtype) is type
        if dtype == str:
            self.dtype = Dataset.STRING_TYPE
        else:
            self.dtype = numpy.dtype(dtype)

        if chunks is None:
            if len(shape) == 0:
                if self.is_string():
                    chunks = (1 * 1024,)
                else:
                    chunks = (8 * 1024,)
            elif len(shape) == 1:
                if self.is_string():
                    n = 1 * 1024 // shape[0]
                else:
                    n = 16 * 1024 // shape[0]
                if n < 2:
                    n = 2
                chunks = (n,) + shape
            elif len(shape) == 2:
                n = 32 * 1024 // shape[0] // shape[1]
                if n < 2:
                    n = 2
                chunks = (n,) + shape
            else:
                raise RuntimeError(f"unsupported shape {shape}")

        self.chunks = None if (chunks is None) else tuple(chunks)
        self.dataset = self.h5file.create_dataset(f"{prefix}/{channel}/{field}", (0,) + self. shape, maxshape=(None,) + self.shape, dtype=self.dtype , chunks=self.chunks, shuffle=shuffle, compression=self.dataset_compression , compression_opts=self.compression_opts)
        self.buf = numpy.zeros(shape=self.chunks, dtype=self.dtype)
        self.nbuf = 0
        self.nwritten = 0

    def append(self, v):
        if self.channel_compression:
            v = decompress(v, self.channel, self.channel_compression, self.shape, self.dtype)
        else:
            if not self.is_value:
                v = numpy.reshape(numpy.frombuffer(v, dtype=self.dtype), self.shape)
        if self.nbuf >= len(self.buf):
            self.flush()
        self.buf[self.nbuf] = v
        self.nbuf += 1

    def flush(self):
        nn = self.nwritten + self.nbuf
        self.dataset.resize((nn,) + self.shape)
        self.dataset[self.nwritten:nn] = self.buf[:self.nbuf]
        self.nwritten = nn
        self.nbuf = 0

    def close(self):
        self.flush()

    def is_string(self):
        return self.dtype == Dataset.STRING_TYPE


class StringDataset(Dataset):
    def __init__(self, prefix, channel, field, h5file, channel_compression, dataset_compression):
        dtype = str
        shape = tuple()
        chunks = (1 * 1024)
        Dataset.__init__(self, prefix, channel, field, h5file, shape, dtype, channel_compression, chunks, dataset_compression=dataset_compression)

class DirectChunkWriteDataset(Dataset):

    def __init__(self, prefix, channel, field, h5file, shape, dtype, channel_compression, dataset_compression):
        shape = tuple(shape)
        chunks = (1,) + shape
        block_size = 0
        compression_opts = (block_size, bitshuffle_compression_lz4)
        shuffle = False
        Dataset.__init__(self, prefix, channel, field, h5file, shape, dtype, channel_compression, chunks, dataset_compression, compression_opts, shuffle)

    def append(self, buf):
        nr = self.nwritten + 1
        self.dataset.resize((nr,)+self.shape)
        #k = struct.unpack(">qi", buf[:12])
        #uncompressed_size = k[0]
        #block_size = k[1]
        #self.compression_opts = (block_size, bitshuffle.h5.H5_COMPRESS_LZ4)
        off = (self.nwritten,) + (0,) * len(self.shape)
        self.dataset.id.write_direct_chunk(off, buf)
        self.nwritten += 1

    def flush(self):
        pass




import inspect

def get_meta(cls, separator=" "):
    signature = inspect.signature(cls)
    pars = signature.parameters
    ret = ""
    for name, par in pars.items():
        if par.kind != inspect.Parameter.VAR_KEYWORD:
            if par.default == inspect.Parameter.empty:
                ret = ret + name + " "
            else:
                if type(par.default) == str:
                    dflt = "'" + par.default + "'"
                else:
                    dflt = str(par.default)
                ret = ret + name + "=" + dflt + separator
    return ret


import redis
import time
from datahub.utils.data import *
import logging
import pickle
_logger = logging.getLogger(__name__)


def on_msg(msg):
    #print(msg)
    pass

def fetch_aligned_data(r, channels, group_name, consumer_name, send_partial_msg=True):
    ids = set()
    group_name = generate_random_string(16)
    for channel in channels:
        try:
            r.xgroup_create(channel, group_name, mkstream=False)
        except Exception as e:
            print(e)
    try:
        sent_id = -1
        streams = {}
        num_channels = len(channels)
        size_align_buffer = 1000
        for channel in channels:
            streams[channel] = ">"
        aligned_data = MaxLenDict(maxlen=size_align_buffer)
        while True:
            entries = r.xreadgroup(group_name, consumer_name, streams, count=10*num_channels, block=200)
            for stream, messages in entries:
                for message_id, message_data in messages:
                    channel = message_data[b'channel'].decode('utf-8')
                    timestamp = int(message_data[b'timestamp'].decode('utf-8'))
                    id = int(message_data[b'id'].decode('utf-8'))
                    ids.add(id)
                    data = message_data[b'value']#.decode('utf-8')
                    value = decode(data)
                    if id not in aligned_data:
                        aligned_data[id] = {}
                        aligned_data[id]["timestamp"] = timestamp
                    aligned_data[id][channel] = value
                    r.xack(stream, group_name, message_id)  # Acknowledge message

            keys_in_order = sorted(list(aligned_data.keys()))
            last_complete_id = -1
            for i in range(len(keys_in_order)-1, 0, -1):
                id = keys_in_order[i]
                val = aligned_data[id]
                complete = len(val) == num_channels + 1
                if complete:
                    last_complete_id = id
                    break
            for i in range(len(keys_in_order)):
                id = keys_in_order[i]
                val = aligned_data[id]
                complete = len(val) == num_channels + 1
                done = False
                if complete:
                    done = True
                elif last_complete_id > id:
                    done = True
                elif len(aligned_data) > (size_align_buffer/2):
                    done = True

                if done:
                    # print ("    - ", id)
                    val = aligned_data.pop(id)
                    if complete or send_partial_msg:
                        if sent_id >= id:
                            print (f"Invalid ID {id} - last sent ID {sent_id}")
                        else:
                           expected = sent_id + 1
                           if (sent_id != -1) and (expected != id):
                               print(f"Missed ID {expected} - received {id}")
                               print(expected in ids, id in ids)
                           msg = collections.OrderedDict()
                           msg["id"] = id
                           msg["timestamp"] = val.get("timestamp", None)
                           for channel in channels:
                               msg[channel] = val.get(channel, None)
                           on_msg(msg)
                           sent_id = id
                    else:
                        logging.debug(f"Discarding partial message: {id}")

            if len(aligned_data) > 10:
                print ("--->", len(aligned_data))
            time.sleep(0.01)
    finally:
        for channel in channels:
            try:
                r.xgroup_destroy(channel, group_name)
            except Exception as e:
                print(e)
if __name__ == '__main__':
    channels = ['channel1', 'channel2', 'channel3']
    with redis.Redis(host='std-daq-build', port=6379, db=0) as r:
        fetch_aligned_data(r, channels, 'mygroup', 'consumer1')



#import redis
#with redis.Redis(host='std-daq-build',port=6379,db=0) as rc:
#    rc.set('test:hello','World')
#    print(rc.get('test:hello'))
#rc.set('test:cnt', 1)
#print(rc.get('test:cnt'))
#rc.lpush("test:l1", "v1", "v2")
#print(rc.lindex('test:l1', 0))
#print(rc.lrange('test:l1', 0, -1))
#print(rc.keys('test:*'))

import unittest
from datahub import *
import time
channels = ['channel1', 'channel2', 'channel3']

class DataBufferTest(unittest.TestCase):

    def test_redis_print(self):
            #with Plot() as plot:
            with Stdout() as stdout:
                with HDF5Writer("/Users/gobbo_a/dev/back/redis.h5") as h5:
                    with Redis(time_type="str") as source:
                        src_ch = source.search("chann");
                        src_db = source.search();
                        source.add_listener(stdout)
                        #source.add_listener(plot)
                        source.add_listener(h5)
                        source.req(channels, 0.0, 2.0)

    def test_redis_dataframe(self):
            with Table() as table:
                with Redis(time_type="str") as source:
                    source.add_listener(table)
                    source.req(channels, 0.0, 5.0)
                    df = table.as_dataframe(index=Table.TIMESTAMP)
                    print(df)
                    df = table.as_dataframe(index=Table.PULSE_ID)
                    print(df)


    def test_redis_stream(self):
        with RedisStream(channels, time_type="str") as source:
            for i in range(10):
                print(i, source.receive(1.0))
        with RedisStream(channels, time_type="str", filter="(channel3>0.5 AND channel1<0.5) OR channel2<0.1") as source:
            for i in range(10):
                print(i, source.receive(1.0))

    def test_redis_stream_as_bsread(self):
        with RedisStream(channels) as source:
            source.forward_bsread(5678)

    def test_redis_stream(self):
        # with RedisStream(channels, time_type="str") as source:
        #    for i in range(1000):
        #        print(i, source.receive(1.0))
        # with RedisStream(channels, time_type="str", filter="(channel3>0.5 AND channel1<0.5) OR channel2<0.1") as source:
        #    for i in range(10):
        #        print(i, source.receive(1.0))
        buf = []
        start = time.time()
        end = start + 6.0

        with RedisStream(channels, time_type="str", queue_size=100, size_align_buffer=1000) as source:
            while time.time() < end:
                rec = source.receive(1.0)
                if rec:
                    buf.append(rec[0])
                    if len(rec[2]) != 3:
                        print(
                            f"Partial message on {time.time() - start} at index {len(buf)} id:{buf[-1]} - keys: {rec[2].keys()}")
                else:
                    print(f"Null message on {time.time() - start} after index {len(buf)} id:{buf[-1] if buf else None}")
            print(f"Number Records: {len(buf)}")

            min_index = buf[0]
            max_index = buf[-1]
            full_range = set(range(min_index, max_index + 1))
            missing_records = sorted(full_range - set(rec for rec in buf))
            missing_indices = [i - min_index for i in missing_records]
            print(f"Number missing: {len(missing_records)}")
            print(f"Missing records: {missing_records}")
            print(f"Missing indices: {missing_indices}")

    def test_redis_merger(self):
        with Redis(time_type="str") as stream1:
            with Redis(time_type="str") as stream2:
                stdout =Stdout()
                merger=Merger(filter = "channel1>0.5")

                stream1.add_listener(merger)
                stream2.add_listener(merger)
                merger.to_source().add_listener(stdout)

                stream1.req(["channel1"], 0.0, 2.0, background=True, filter = "channel1>0.5")
                stream2.req(["channel2"], 0.0, 2.0, background=True)
                stream1.join()
                stream2.join()


if __name__ == '__main__':
    unittest.main()



from datahub import *

TIMEOUT = None
_logger = logging.getLogger(__name__)

def validate_response(server_response):
    if server_response["state"] != "ok":
        raise ValueError(server_response.get("status", "Unknown error occurred."))
    return server_response

def get_response(url, post=None, params=None):
    import requests
    if post:
        server_response = requests.post(url, json=None if (post==True) else post, params=params, timeout=TIMEOUT).json()
    else:
        server_response = requests.get(url, params=params, timeout=TIMEOUT).json()
    return validate_response(server_response)


def split_suffix_in_brackets(s):
    if not s:
        return None, None
    pattern = re.compile(r"(.*)\[(.*)\]$")
    match = pattern.search(s)
    if match:
        # Extract the prefix and the value inside brackets
        prefix = match.group(1)
        value = match.group(2)
        if not prefix:
            prefix=None
        return prefix, value
    else:
        return s, None

class Pipeline(Bsread):
    """
    Retrieves data from CamServer pipelines.
    """
    DEFAULT_URL = os.environ.get("PIPELINE_DEFAULT_URL", "http://sf-daqsync-01:8889")

    def __init__(self, url=DEFAULT_URL, name=None, config=None, mode="SUB", path=None, **kwargs):
        """
        url (str, optional): PipelineServer URL. Default value can be set by the env var PIPELINE_DEFAULT_URL.
        name (str, optional): name of the instance and/or pipeline in the format: INSTANCE_NAME[PIPELINE_NAME]
        config (dict, optional): pipeline configuration (or additional configuration if pipeline name is defined)
        mode (str, optional): "SUB" or "PULL"
        path (str, optional): hint for the source location in storage or displaying.
        """
        self.address = url
        if name or config:
            instance, pipeline = split_suffix_in_brackets(name)
            try:
                if not instance:
                    raise Exception("Create")
                url = self.get_stream(instance)
            except:
                if not config:
                    if pipeline:
                        url = self.create_instance_from_name(name=pipeline, instance_id=instance)
                else:
                    if type(config)==str:
                        config = eval(config)
                    if pipeline:
                        url = self.create_instance_from_name(name=pipeline, instance_id=instance, additional_config=config)
                    else:
                        url = self.create_stream_from_config(config, instance)

        Bsread.__init__(self, url=url, mode=mode, path=path, name=name, **kwargs)

    def get_stream(self, instance_id):
        rest_endpoint = "/api/v1/pipeline/instance/%s" % instance_id
        return get_response(self.address+rest_endpoint)["stream"]

    def create_stream_from_config(self, config, instance_id=None):
        params=None
        if instance_id:
            params = {"instance_id": instance_id} if instance_id else None
            rest_endpoint = "/api/v1/pipeline"
        else:
            rest_endpoint = "/api/v1/pipeline/instance/"
        return get_response(self.address+rest_endpoint, post=config, params=params)["stream"]

    def create_instance_from_name(self, name, instance_id=None, additional_config = None):
        params=None
        rest_endpoint = "/api/v1/pipeline/" + name
        if instance_id or additional_config:
            params = {}
            if instance_id:
                params["instance_id"] = instance_id
            if additional_config:
                #params["additional_config"] = additional_config if type(additional_config) is str else json.dumps(additional_config)
                params["additional_config"] = json.dumps(additional_config)
        return get_response(self.address+rest_endpoint, post=True, params=params)["stream"]


    def get_instances(self):
        rest_endpoint = "/api/v1/pipeline"
        return get_response(self.address+rest_endpoint)["pipelines"]

    def search(self, regex):
        ret = self.get_instances()
        print(ret)
        if regex:
            ret = [element for element in ret if regex in element]
        pd = self._get_pandas()
        if pd:
            if len(ret) == 0:
                return None
            df = pd.DataFrame(ret, columns=["instances"])
            ret = df.to_string(index=False)
        return ret



class Camera(Bsread):
    """
    Retrieves data from CamServer cameras.
    """
    DEFAULT_URL = os.environ.get("CAMERA_DEFAULT_URL", "http://sf-daqsync-01:8888")

    def __init__(self, url=DEFAULT_URL, name=None, mode="SUB", path=None, **kwargs):
        """
        url (str, optional): CameraServer URL. Default value can be set by the env var CAMERA_DEFAULT_URL.
        name (str): camera name
        mode (str, optional): "SUB" or "PULL"
        path (str, optional): hint for the source location in storage or displaying.
        """
        self.address = url
        if name:
            url = self.get_instance_stream(name)
        Bsread.__init__(self, url=url, mode=mode, path=path, name=name, **kwargs)


    def get_instance_stream(self, instance_id):
        rest_endpoint = "/api/v1/cam/%s" % instance_id
        return get_response(self.address+rest_endpoint)["stream"]

    def get_instances(self):
        rest_endpoint = "/api/v1/cam"
        return get_response(self.address+rest_endpoint)["cameras"]

    def search(self, regex):
        ret = self.get_instances()
        if regex:
            ret = [element for element in ret if regex in element]
        pd = self._get_pandas()
        if pd:
            if len(ret) == 0:
                return None
            df = pd.DataFrame(ret, columns=["instances"])
            ret = df.to_string(index=False)
        return ret






try:
    import epics
except:
    epics = None

from datahub import *

GENERATE_ID = False

_logger = logging.getLogger(__name__)

class Channel:
    def __init__(self, name, source):
        if epics is None:
            raise Exception("EPICS library not available")
        self.name = name
        self.channel = epics.PV(name, auto_monitor=True)
        self.id = 0
        self.source = source
        # channel.wait_for_connection(config.EPICS_TIMEOUT)

    def start(self):
        def callback(value, timestamp, status, **kwargs):
            channel_name =self.name
            timestamp = create_timestamp(timestamp)
            self.source.receive_channel(self.name, value, timestamp, self.pulse_id if GENERATE_ID else None, check_types=True)
            self.id = self.id + 1
        self.channel.add_callback(callback)

    def stop(self):
        self.channel.clear_callbacks()


    def close(self):
        try:
            self.channel.disconnect()
        except:
            pass

class Epics(Source):
    """
    Retrieves data from the EPICS channels.
    """
    DEFAULT_URL = None
    def __init__(self, url=DEFAULT_URL, path=None, **kwargs):
        """
        url (str, optional): if defined sets the EPICS_CA_ADDR_LIST.
        path (str, optional): hint for the source location in storage or displaying.
        """
        Source.__init__(self, url=url, path=path, **kwargs)
        if self.url:
            os.environ["EPICS_CA_ADDR_LIST"] = self.url

    def run(self, query):
        channels_names = query.get("channels", [])
        channels = []
        for name in channels_names:
            channel = Channel(name, self)
            channels.append(channel)
        try:
            self.range.wait_start()
            for channel in channels:
                channel.start()
            while self.range.is_running() and not self.aborted:
                time.sleep(0.1)
            for channel in channels:
                channel.stop()
            self.close_channels()
        finally:
            for channel in channels:
                channel.close()


    def search(self, pattern):
        #If backend given that tread as facility, Otherwise search the environment
        import requests
        facility = self.get_backend()
        if not facility:
            facility = os.environ.get("SYSDB_ENV", None)
        api_base_address = "http://iocinfo.psi.ch/api/v2"
        pattern = ".*" + pattern + ".*"  #No regex
        parameters = {"pattern": pattern}
        if facility:
            parameters["facility"] = facility
        parameters["limit"] = 0
        response = requests.get(api_base_address + "/records", params=parameters)
        response.raise_for_status()
        channels = response.json()
        ret = channels
        if not self.verbose:
            pd = self._get_pandas()
            if pd is None:
                if facility:
                    ret = [record["name"] for record in channels]
                else:
                    ret = [f"{record['name']} [{record['facility']}]" for record in channels]
            else:
                ret = None
                header = list(channels[0].keys()) if len(channels) > 0 else []
                data = [d.values() for d in channels]
                df = pd.DataFrame(data, columns=header)
                if (len(header) > 0):
                    df = df.sort_values(by=[header[0], header[1]])
                    columns_to_display = ["facility", "name", "type", "ioc"] #, "description"]
                    ret = df[columns_to_display].to_string(index=False)
        return ret



from datahub import *
try:
    import redis
except:
    redis = None

class Stddaq(Bsread):
    """
    Retrieves data from CamServer cameras.
    """
    DEFAULT_URL = os.environ.get("STDDAQ_DEFAULT_URL", "sf-daq-6.psi.ch:6379")

    def __init__(self, url=DEFAULT_URL, name=None, mode="SUB", path=None, **kwargs):
        """
        url (str, optional): URL for Stddaq Redis repo.
        name (str): device name
        mode (str, optional): "SUB" or "PULL"
        path (str, optional): hint for the source location in storage or displaying.
        """
        self.address = url
        if redis is None:
            raise Exception("PyRedis library not available")
        self.host, self.port = get_host_port_from_stream_address(url)
        self.db='0'
        if name:
            url = self.get_instance_stream(name)
        Bsread.__init__(self, url=url, mode=mode, path=path, name=name, **kwargs)

    def get_instance_stream(self, name):
        with redis.Redis(host=self.host, port=self.port, db=self.db) as r:
            ret = r.get(name)
            return ret.decode('utf-8').strip()   if ret else ret


    def search(self, regex):
        with redis.Redis(host=self.host, port=self.port, db=self.db) as r:
            if not regex:
                #return r.config_get('databases')
                return r.info('keyspace')
            else:
                cursor = '0'
                streams = []
                match = f'*{regex}*' if regex else '*'
                while cursor != 0:
                    cursor, keys = r.scan(cursor=cursor, match=match)
                    for key in keys:
                        if r.type(key) == b'string':
                            if type(key) != str:
                                key = key.decode('utf-8')
                            streams.append(key)
                return sorted(streams)



import unittest
from datahub import *


class TimingTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_pulse_id(self):
        t1 = time.time()
        id = time_to_pulse_id(t1)
        t2 = pulse_id_to_time(id)
        print (t1-t2)
        self.assertLess(abs(t1-t2), PULSE_ID_INTERVAL)

    def test_performance(self):
            # Comparimg performance for multiplying  by 100 and dividing by 0.01
            import timeit
            time_multiply_int = timeit.timeit('result = x * 100', setup='x = 1.23456789', number=1000000)
            time_multiply = timeit.timeit('result = x * 100.0', setup='x = 1.23456789', number=1000000)
            time_divide = timeit.timeit('result = x / 0.01', setup='x = 1.23456789', number=1000000)
            print(f'Time taken for int multiplication: {time_multiply_int:.6f} seconds')
            print(f'Time taken for multiplication: {time_multiply:.6f} seconds')
            print(f'Time taken for division: {time_divide:.6f} seconds')

if __name__ == '__main__':
    unittest.main()



import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import pandas as pd
from datahub import *
from datetime import date, datetime, timedelta
from dateutil.relativedelta import relativedelta
import pandas as pd
import dash
#import dash_bootstrap_components as dbc
from dash import Dash, html, dcc, callback, Output, Input, State, dcc
from dash.exceptions import PreventUpdate
import plotly.graph_objects as go
backend = "sf-databuffer"
time_fmt = "%Y-%m-%d %H:%M:%S"
# Initialize the Dash app
app = dash.Dash(__name__)

# Dummy function to simulate receiving a dataset from a server
def get_dataset():
    source = Daqbuf(backend=backend, cbor=True, parallel=True, time_type="seconds")
    table = Table()
    source.add_listener(table)
    query = {"channels": ["S10BC01-DBPM010:Q1", "S10BC01-DBPM010:X1"],
         "start": (datetime.now() - timedelta(hours=1)).strftime(time_fmt),
         "end": datetime.now().strftime(time_fmt), "bins": 100}
    source.request(query)
    df = table.as_dataframe(Table.TIMESTAMP)
    if df is not None:
        df.reindex(sorted(df.columns), axis=1)
    return df


# Layout of the Dash app
app.layout = html.Div([
    dcc.Store(id='stored-data'),
    dcc.Dropdown(
        id="dropdown",
        options=[
            {"label": "Option 1", "value": "option1"},
            {"label": "Option 2", "value": "option2"}
        ],
        value="option1"
    ),
    html.Button('Load Data', id='load-data-button', n_clicks=0),
    html.Div(id="table-container")
])

# Callback to load data and store it in dcc.Store
@app.callback(
    Output('stored-data', 'data'),
    Input('load-data-button', 'n_clicks')
)
def load_data(n_clicks):
    if n_clicks > 0:
        dataset = get_dataset()
        return dataset.to_dict()
    return {}

# Callback to update the table based on dropdown selection
@app.callback(
    Output("table-container", "children"),
    [Input("dropdown", "value"), Input('stored-data', 'data')]
)
def update_table(selected_option, stored_data):
    if stored_data:
        dataset = pd.DataFrame(stored_data)
        if selected_option == "option1":
            return html.Table([
                html.Tr([html.Th(col) for col in dataset.columns]),
                html.Tr([html.Td(dataset.iloc[0][col]) for col in dataset.columns])
            ])
        else:
            return html.Table([
                html.Tr([html.Th(col) for col in dataset.columns]),
                html.Tr([html.Td(dataset.iloc[1][col]) for col in dataset.columns])
            ])
    return "No data loaded yet."

# Run the app
if __name__ == "__main__":
    app.run_server(debug=True)



########################################################################################################################
# Compression Utilities
########################################################################################################################

import numpy
import struct
import logging

_logger = logging.getLogger(__name__)

class Endianness:
     LITTLE = "LITTLE_ENDIAN"
     BIG = "BIG_ENDIAN"

try:
    import bitshuffle
    import bitshuffle.h5
    bitshuffle_compression_filter =  bitshuffle.h5.H5FILTER
    bitshuffle_compression_lz4 = bitshuffle.h5.H5_COMPRESS_LZ4

except:
    _logger.error("bitshuffle not installed: BITSHUFFLE_LZ4 compression not supported")
    #itshuffle does not need to be present to dump compressed channels
    bitshuffle = None
    bitshuffle_compression_filter =32008
    bitshuffle_compression_lz4 = 2

class Compression:
    BITSHUFFLE_LZ4 = bitshuffle_compression_filter
    GZIP = "gzip"
    SZIP = "szip"
    LZF = "lzf"

def decompress(blob, name, compression, shape, dtype, border=Endianness.LITTLE):
    if bitshuffle is None:
        raise Exception("Bitshuffle not available")
    if compression == Compression.BITSHUFFLE_LZ4:
        c_length = struct.unpack(">q", blob[0:8])[0]
        b_size = struct.unpack(">i", blob[8:12])[0]
        nbuf = numpy.frombuffer(blob[12:], dtype=numpy.uint8)
        if dtype == "str":
            if len(shape) > 0:
                raise RuntimeError(f"Compression of arrays of strings not supported")
            #if c_length < 1 or c_length > 4 * 1024:
            #    raise RuntimeError(f"unexpected string size: {c_length}")
            #if b_size < 512 or b_size > 16 * 1024:
            #    raise RuntimeError(f"unexpected block size: {b_size}")
            dtype = numpy.dtype(numpy.int8) #numpy.dtype("b")
            value = bitshuffle.decompress_lz4(nbuf, shape=(c_length,), dtype=dtype, block_size=b_size)
            value = value.tobytes().decode()
        else:
            if len(shape) == 0:
                raise RuntimeError(f"Compression not supported on scalar numeric data {name}  shape {shape}  dtype {dtype}")
            #if c_length < 1 or c_length > 1 * 1024 * 1024:
            #    raise RuntimeError(f"unexpected value size: {c_length}")
            #if b_size < 512 or b_size > 16 * 1024:
            #    raise RuntimeError(f"unexpected block size: {b_size}")
            if type(dtype) == str:
                dtype = numpy.dtype(dtype).newbyteorder('<' if border == "LITTLE_ENDIAN" else ">")
            block_size = int(b_size / dtype.itemsize) #0
            value = bitshuffle.decompress_lz4(nbuf, shape=shape, dtype=dtype ,block_size=block_size)
    else:
        raise RuntimeError(f"Compression type {compression} is not supported.")
    return value




try:
    import redis
except:
    redis = None

from datahub import *
from datahub.utils.align import *

import threading

_logger = logging.getLogger(__name__)

class Redis(Source):
    """
    Retrieves data from the Redis or Dragonfly streams.
    """

    DEFAULT_URL = os.environ.get("REDIS_DEFAULT_URL", 'sf-daqsync-18:6379')
    DEFAULT_BACKEND = os.environ.get("REDIS_DEFAULT_BACKEND", '0')

    def __init__(self, url=DEFAULT_URL, backend=DEFAULT_BACKEND, path=None, **kwargs):
        """
        url (str, optional): Redis URL. Default value can be set by the env var REDIS_DEFAULT_URL.
        backend (str): Redis database. Default value can be set by the env var REDIS_DEFAULT_BACKEND.
        path (str, optional): hint for the source location in storage or displaying.
        """
        Source.__init__(self, url=url, backend=backend, path=path, **kwargs)
        if redis is None:
            raise Exception("PyRedis library not available")
        self.consumer_name = 'datahub'
        self.host, self.port = get_host_port_from_stream_address(self.url)
        self.db = self.backend
        self.messages = []


    def create_group(self, r, channels):
        group_name = generate_random_string(16)
        try:
            pipeline = r.pipeline()
            for channel in channels:
                pipeline.xgroup_create(channel, group_name, mkstream=False)
            pipeline.execute()
        except Exception as e:
            _logger.warning(f"Error creating stream group: {str(e)}")

        return group_name

    def destroy_group(self, r, channels, group_name):
        try:
            pipeline = r.pipeline()
            for channel in channels:
                pipeline.xgroup_destroy(channel, group_name)
            pipeline.execute()
        except Exception as e:
            _logger.warning(f"Error destroying stream group: {str(e)}")

    def run(self, query):
        partial_msg = query.get("partial_msg", True)
        utc_timestamp = query.get("utc_timestamp", True)
        channels = query.get("channels", [])
        size_buffer = query.get("size_buffer", 1000)
        filter = query.get("filter", None)
        align = Align(self.on_msg, channels, self.range, filter , partial_msg=partial_msg, size_buffer=size_buffer, utc_timestamp=utc_timestamp)

        with redis.Redis(host=self.host, port=self.port, db=self.db, decode_responses=False) as r:
            group_name = self.create_group(r, channels)
            try:
                streams = {channel : ">" for channel in channels}
                while not self.range.has_ended() and not self.aborted:
                    entries = r.xreadgroup(group_name, self.consumer_name, streams, count=5 * len(channels), block=100)
                    if entries:
                        for stream, messages in entries:
                            processed_ids = []
                            try:
                                for message_id, message_data in messages:
                                    channel = message_data[b'channel'].decode('utf-8')
                                    timestamp = int(message_data[b'timestamp'].decode('utf-8'))
                                    id = int(message_data[b'id'].decode('utf-8'))
                                    data = message_data[b'value']
                                    value = decode(data)
                                    align.add(id, timestamp, channel, value)
                                    processed_ids.append(message_id)
                            finally:
                                r.xack(stream, group_name, *processed_ids)
                        align.process()
            finally:
                self.destroy_group(r, channels, group_name)
                self.close_channels()

    def on_msg(self, id, timestamp, msg):
        for channel_name in msg.keys():
            v = msg.get(channel_name, None)
            if v is not None:
                self.receive_channel(channel_name, v, timestamp, id, check_changes=True, check_types=True)

    def close(self):
        Source.close(self)

    def search(self, regex=None):
        with redis.Redis(host=self.host, port=self.port, db=self.db, decode_responses=True) as r:
            if not regex:
                #return r.config_get('databases')
                return r.info('keyspace')
            else:
                cursor = '0'
                streams = []
                match = f'*{regex}*' if regex else '*'
                while cursor != 0:
                    cursor, keys = r.scan(cursor=cursor, match=match)
                    for key in keys:
                        if regex in key and r.type(key) == 'stream':
                                if type(key)!=str:
                                    key = key.decode('utf-8')
                                streams.append(key)
                return sorted(streams)

class RedisStream(Redis):

    def __init__(self, channels, filter=None, queue_size=100,  **kwargs):
        Redis.__init__(self, **kwargs)
        self.message_buffer = collections.deque(maxlen=queue_size)
        self.condition = threading.Condition()
        self.req(channels, 0.0, 365 * 24 * 60 * 60, filter=filter, background=True)

    def close(self):
        Redis.close(self)

    def on_msg(self, id, timestamp, msg):
        with self.condition:
            self.message_buffer.append((id, timestamp, msg))
            self.condition.notify()

    def drain(self):
        with self.condition:
            self.message_buffer.clear()

    def receive(self, timeout=None):
        with self.condition:
            if not self.message_buffer:
                self.condition.wait(timeout)
            if self.message_buffer:
                return self.message_buffer.popleft()

    def forward_bsread(self, port, mode="PUB"):
        from datahub.utils.bsread import create_sender
        sender = create_sender(port, mode)
        try:
            while True:
                id, timestamp, data = self.receive()
                timestamp = (int(timestamp / 1e9), int(timestamp % 1e9)) #float(timestamp) / 1e9
                sender.send(data=data, pulse_id=id, timestamp=timestamp, check_data=True)
        finally:
            sender.close()



from bsread.sender import Sender, BIND, CONNECT
from bsread import PUB, SUB, PUSH, PULL


def create_sender(port, mode, block=False, queue_size=10, compression=None):
    if mode == "PUB":
        mode = PUB
    if mode == "PUSH":
        mode = PUSH
    sender = Sender(port=port,
                address="tcp://*",
                conn_type=BIND,
                mode=mode,
                queue_size=queue_size,
                block=block,
                data_compression=compression,
                data_header_compression=None)
    sender.open()
    return sender


from datahub import *
import unittest
from datahub import *

class EpicsTest(unittest.TestCase):

    def test_databuffer(self):
        query = {
            "channels": ["S10BC01-DBPM010:Q1", "S10BC01-DBPM010:X1"],
            "start": "2024-02-14 08:50:00.000",
            "end": "2024-02-14 08:50:05.000"
        }

        with DataBuffer(backend="sf-databuffer") as source:
            table = Table()
            source.add_listener(table)
            source.request(query)
            dataframe = table.as_dataframe(Table.PULSE_ID)
            print(dataframe)

    def test_epics(self):
        query = {
            "channels": ["TESTIOC:TESTCALCOUT:Input", "TESTIOC:TESTSINUS:SinCalc"],
            "start": "0.0",
            "end": "2.0",
        }
        with Epics(url="localhost:54321", time_type="str") as source:
            table = Table()
            source.add_listener(table)
            source.request(query)
            dataframe = table.as_dataframe(Table.TIMESTAMP)
            print(dataframe)



import logging
from datahub import Consumer

_logger = logging.getLogger(__name__)


class Table(Consumer):
    TIMESTAMP = "timestamp"
    PULSE_ID = "pulse_id"
    def __init__(self, **kwargs):
        Consumer.__init__(self, **kwargs)
        self.clear()

    def clear(self):
        self.data = {}

    def on_close(self):
        self.clear()

    def on_start(self, source):
        pass

    def on_stop(self, source, exception):
        pass

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        self.data[name] = []
        if (metadata.get("bins", None)):
            for col in "max", "min", "count":
                self.data[f"{name} {col}"] = []

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        self.data[name].append({Table.TIMESTAMP: timestamp, Table.PULSE_ID: pulse_id, name: value})
        if kwargs.get("bins", None):
            for col in "max", "min", "count":
                arg_name = f"{name} {col}"
                self.data[arg_name].append({Table.TIMESTAMP: timestamp, Table.PULSE_ID: pulse_id, arg_name: kwargs[col]})

    def on_channel_completed(self, source, name):
        pass

    def as_dataframe(self, index=TIMESTAMP, replace_nan=False):
        import pandas as pd
        dataframe = None
        drop = Table.PULSE_ID if index!=Table.PULSE_ID else Table.TIMESTAMP
        for key in self.data:
            values = self.data[key]
            if values is not None and (len(values) > 0):
                df = pd.DataFrame(self.data[key])
                df = df.drop(columns=[drop])
                df = df.set_index(index)
                if dataframe is None:
                    dataframe = df
                else:
                    dataframe = pd.merge(dataframe, df, how='outer', on=index)
        if (dataframe is not None) and replace_nan:
            #Fill NA/NaN values by propagating the last valid observation to next valid.
            dataframe.ffill(inplace=True)
        if (dataframe is not None):
            dataframe.sort_index(inplace=True)
        return dataframe



import unittest
from datahub import *
import time


class DataBufferTest(unittest.TestCase):

    def test_listeners(self):
        with Plot() as plot:
            with Bsread() as source:
                source.add_listener(plot)
                source.req(["S10BC01-DBPM010:X1", "S10BC01-DBPM010:Q1"], 0.0, 2.0)


if __name__ == '__main__':
    unittest.main()



from datahub import *

start = "2024-06-14 09:00:00"
end = "2024-06-14 09:01:00"
backend = "sf-databuffer"





with Daqbuf(backend=backend, cbor=True, parallel=True, time_type="str") as stream1:
    with Daqbuf(backend=backend, cbor=True, parallel=True, time_type="str") as stream2:
        stdout =Stdout()
        merger=Merger(filter = "S10BC01-DBPM010:X1>0")

        stream1.add_listener(merger)
        stream2.add_listener(merger)

        merger.to_source().add_listener(stdout)

        stream1.req(["S10BC01-DBPM010:Q1"], start, end, background=True)
        stream2.req(["S10BC01-DBPM010:X1"], start, end, background=True)
        stream1.join()
        stream2.join()



import unittest
from datahub import *


backend = "sf-imagebuffer"
#url = "https://data-api.psi.ch/api/1"
url = "http://sf-daq-5.psi.ch:8380/api/1"
#url = "http://127.0.0.1:7777/api/1"

#url = "https://data-api.psi.ch/api/1"
#url = "http://sf-daq-5.psi.ch:8380/api/1"
#url = "http://sf-daq-5.psi.ch/api/1"

filename = "ib.h5"
channels = ["SATES21-CAMS154-M2:FPICTURE"]
channels = ["SARFE10-PSSS059:FPICTURE"]
start = "2024-02-20 12:04:31.000"
end = "2024-02-20 12:04:32.000"

query = {
    "channels": channels,
    "start": start,
    "end": end
}


class RetrivalTest(unittest.TestCase):
    def setUp(self):
        self.source = Retrieval(url=url, backend=backend, auto_decompress=False)

    def tearDown(self):
        self.source.close()


    def test_listeners(self):
        hdf5 = HDF5Writer(filename)
        #stdout = Stdout()
        #table = Table()
        self.source.add_listener(hdf5)
        #self.source.add_listener(stdout)
        #self.source.add_listener(table)
        self.source.request(query)
        #dataframe = table.as_dataframe()
        #self.assertEqual(list(dataframe.keys()), channels)
        self.source.close_listeners()


if __name__ == '__main__':
    unittest.main()


from datahub import Consumer, Source
from datahub.utils.align import *

class Merger (Consumer):
    def __init__(self,  callback=None, filter=None, partial_msg=True, **kwargs):
        Consumer.__init__(self, **kwargs)
        self.channels = {}
        self.align = Align(self.on_received_message, None, range=None, filter=filter, partial_msg=partial_msg)
        self.callback = callback

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        if self.align:
            self.align.add(pulse_id, timestamp, name, value)
            self.align.process()

    def on_start(self, source):
        self.channels[source] = source.query.get("channels",[])
        if len(self.channels) >1:
            channels = []
            for source, ch in self.channels.items():
                channels = channels + ch
            self.align.set_channels(channels)

    def on_stop(self, source, exception):
        del self.channels[source]
        if len(self.channels) == 0:
            self.align.set_channels(None)
            self.align.reset()

    def on_received_message(self, id, timestamp, msg):
        if self.callback:
            self.callback(id, timestamp, msg)

    def to_source(self):
        class MergerSource (Source):
            def __init__(self, merger, **kwargs):
                Source.__init__(self, **kwargs)
                self.merger = merger;
                self.merger.callback = self.on_received_message

            def on_received_message(self, id, timestamp, msg):
                for channel, value in msg.items():
                    self.receive_channel(channel, value, timestamp, id, check_changes=True)
        return MergerSource(self)


import time
import traceback

from datahub import *
from datetime import date, datetime, timedelta
from dateutil.relativedelta import relativedelta
import pandas as pd
import dash
#import dash_bootstrap_components as dbc
from dash import Dash, html, dcc, callback, Output, Input, State, dcc
from dash.exceptions import PreventUpdate
import plotly.graph_objects as go
import flask
from flask import session
import traceback
import numpy as np

backend = "sf-databuffer"
time_fmt = "%Y-%m-%d %H:%M:%S"
colors = [ (0 ,0, 255),  (255, 0, 0),  (0 ,255, 0), (0 ,127, 127),  (127, 127, 0),  (127 ,0, 127), (127 ,127, 127) ]


def search_channels(regex):
    if not regex:
        return []
    try:
        with  Daqbuf(backend=backend) as source:
            source.verbose = True
            ret = source.search(regex)
            results = [ch["name"] for ch in  ret.get("channels", [])]
    except:
        results = []

    return [{'label': result, 'value': result} for result in results]

def fetch_data(bins, channels, start, end):
    query = {
        "channels": channels,
        "start": start,
        "end": end,
    }
    if (bins):
        query["bins"] = bins
    session["query"] = query

    with Daqbuf(backend=backend, cbor=True, parallel=True, time_type="seconds") as source:
        table = Table()
        source.add_listener(table)
        source.request(query)
        df = table.as_dataframe(Table.TIMESTAMP)
        if df is not None:
            df.reindex(sorted(df.columns), axis=1)
        #session["df"]=df.to_dict()
        return df


def get_series(df, channel, color, index):
    cols = list(df.keys())
    binned = (channel + " max") in cols
    x = pd.to_datetime(df.index, unit='s')
    y = df[channel]

    return go.Scatter(
        x=x,
        y=y,
        mode='lines+markers',
        name=channel,
        line=dict(color=f'rgba({color[0]}, {color[1]}, {color[2]}, 1.0)'),
        marker=dict(
            color=f'rgba({color[0]}, {color[1]}, {color[2]}, 1.0)',
            size=5
        ),
        #yaxis = "y" if (index==0) else ("y"+str(index+1)),
        error_y=None if not binned else dict(
            type='data',
            symmetric=False,
            array=df[channel + " max"] - y,
            arrayminus=y - df[channel + " min"],
            # array= [1.0] * len(df.index),
            # arrayminus= [2.0] * len(df.index),
            visible=True,
            color=f'rgba({color[0]}, {color[1]}, {color[2]}, 0.1)'  # Set the color with transparency
        )
    )

def init_session():
    if session.get("query", None) is None:
        session["query"] = ""
        #session["source"] = source
        #session["table"] = table
        #session["df"] = source


def get_figure(df, channels):
    if (df is None) or (channels is None) :
        return {}
    data = []
    index = 0
    for channel in channels:
        if (channel in df.keys()):
            data.append(get_series(df, channel, colors[index % len(colors)], index))
            index+=1
    return {
        'data': data,
        'layout': go.Layout(
            #title=str(query),
            xaxis={'title': None},
            yaxis={'title': channels[0] if len(channels)==1 else None},
            margin=dict(t=40, b=40, l=80, r=40)  # Adjust the top margin (t) to make it smaller
        )
    }

def fetch_graphs(df, single):
    if single:
        return [
            dcc.Graph(id='graph', figure= get_figure(df, session["query"]["channels"]))
        ]
    else:
        return [
            dcc.Graph(
                id='graph_' + channel, figure=get_figure(df, [channel])
            ) for channel in session["query"]["channels"]
        ]


# Create the Dash app
server = flask.Flask(__name__)
app = dash.Dash(__name__, server=server, title='Daqbuf UI', update_title='Loading...',suppress_callback_exceptions=True)
server.config.update(SECRET_KEY=os.urandom(24))

cmp_height = '30px'
label_style = {'margin-right': '0px', 'minHeight': cmp_height, 'display': 'flex', 'alignItems': 'center'}
input_style = {'margin-right': '10px', 'textAlign': 'center', 'minHeight': cmp_height}
check_style={'margin-right': '10px','minWidth': '80px', 'minHeight': cmp_height, 'display': 'flex', 'alignItems': 'center'}
drop_style = {'margin-right': '10px', 'min-width': '130px', 'minHeight': cmp_height}
range_options = ["Last 1min", "Last 10min", "Last 1h", "Last 12h", "Last 24h", "Last 7d", "Yesterday", "Today", "Last Week", "This Week", "Last Month", "This Month"]

def _getCheckSingleOptions(disabled=False):
    return [{'label': 'Single', 'value': 'single', 'disabled': disabled}]

# Define the layout of the app
app.layout = html.Div(children=[
    html.H1(children='Daqbuf UI'),
    html.Div([
        html.Label('Bins:', style=label_style),
        dcc.Input(id='input_bins', type='number', min=0, max=1000, step=1, style=input_style, value=100),
        html.Label('From:', style=label_style),
        dcc.Input(id='input_from', type='text', style=input_style, value=""),
        html.Label('To:', style=label_style),
        dcc.Input(id='input_to', type='text', style=input_style, value=""),
        dcc.Dropdown(id='dropdown_set_range', placeholder='Set time range', style=drop_style, options=range_options),
        html.Label('Channels:', style=label_style),
        dcc.Dropdown(id='dropdown_channels', placeholder='Enter query channel names', multi=True, value=[], options=[], style={'margin-right': '20px',  'minWidth': '100px', 'width':"100%", 'minHeight': cmp_height}),
        dcc.Checklist(id='checkbox_single',options=_getCheckSingleOptions(), style=check_style),
        html.Button('Query', n_clicks=0, id='button_query', style={'minWidth': '100px', 'minHeight': cmp_height}),
        dcc.Loading(id="loading",type="default",children=html.Div(id='loading_output')),
        dcc.Store(id='store', data=''),
    ], style={'display': 'flex', 'align-items': 'center', 'margin-bottom': '10px', 'justifyContent': 'center', }),
    html.Div(id='output-container', style={'margin-top': '20px'})
])

@app.callback(
    Output('dropdown_set_range', 'value'),
    Input('store', 'data')
)
def on_page_load(data):
    session["init"]=True
    return ""

@app.callback(
    Output('input_from', 'value', allow_duplicate=True),
    Output('input_to', 'value', allow_duplicate=True),
    Input('dropdown_set_range', 'value'),
    prevent_initial_call=True
)
def set_range(value):
    if not value:
        if session.get("init", False):
            value = range_options[2]
            session["init"] = False
        else:
            raise PreventUpdate

    now = datetime.now()
    start = None
    end = now

    if value == range_options[0]:
        start = now - timedelta(minutes=1)
    elif value == range_options[1]:
        start = now - timedelta(minutes=10)
    elif value == range_options[2]:
        start = now - timedelta(hours=1)
    elif value == range_options[3]:
        start = now - timedelta(hours=12)
    elif value == range_options[4]:
        start = now - timedelta(hours=24)
    elif value == range_options[5]:
        start = now - timedelta(days=7)
    elif value == range_options[6]:
        yesterday_date = now.date() - timedelta(days=1)
        start = datetime.combine(yesterday_date, datetime.min.time())
        end = datetime.combine(yesterday_date, datetime.max.time())
    elif value == range_options[7]:
        start = datetime.combine(now.date(), datetime.min.time())
    elif value == range_options[8]:
        start_of_current_week = now.date() - timedelta(days=now.weekday())
        end_of_last_week = start_of_current_week - timedelta(days=1)
        start_of_last_week = end_of_last_week - timedelta(days=6)
        start = datetime.combine(start_of_last_week, datetime.min.time())
        end = datetime.combine(end_of_last_week, datetime.max.time())
    elif value == range_options[9]:
        start = datetime.combine(now.date() - timedelta(days=now.weekday()), datetime.min.time())
    elif value == range_options[10]:
        previous_month = now - relativedelta(months=1)
        first_day_of_previous_month = previous_month.replace(day=1)
        last_day_of_previous_month = previous_month.replace(day=1) + relativedelta(months=1) - timedelta(days=1)
        start = datetime.combine(first_day_of_previous_month, datetime.min.time())
        end = datetime.combine(last_day_of_previous_month, datetime.max.time())
    elif value == range_options[11]:
        first_day_of_current_month = now.replace(day=1)
        start = datetime.combine(first_day_of_current_month, datetime.min.time())
    start = start.strftime(time_fmt)
    end = end.strftime(time_fmt)
    return start, end


@callback(
    Output('button_query', 'disabled', allow_duplicate=True),
    Output('checkbox_single', 'options', allow_duplicate=True),
    Input('checkbox_single', 'value'),
    prevent_initial_call=True
)
def on_click_single(value):
    return True, _getCheckSingleOptions(True)

@callback(
    Output('button_query', 'disabled', allow_duplicate=True),
    Output('checkbox_single', 'options', allow_duplicate=True),
    Input('button_query', 'n_clicks'),
    prevent_initial_call=True
)
def on_click(n_clicks):
    if not n_clicks:
        raise PreventUpdate
    return True, _getCheckSingleOptions(True)

@callback(
    Output('output-container', 'children'),
    Output('loading_output', 'children'),
    Output('button_query', 'disabled', allow_duplicate=True),
    Output('checkbox_single', 'options', allow_duplicate=True),
    Input('button_query', 'n_clicks'),
    Input('checkbox_single', 'value'),
    State('input_bins', 'value'),
    State('dropdown_channels', 'value'),
    State('input_from', 'value'),
    State('input_to', 'value'),
    State('checkbox_single', 'value'),
    State('button_query', 'disabled'),
    prevent_initial_call=True
)

def update_data(n_clicks, _, bins, channels, start, end, single, querying):
    init_session()
    graphs = ''
    single = single[0] if single else None
    if n_clicks:
        if len(channels)>0:
            try:
                df = fetch_data(bins, channels, start, end)
                graphs = fetch_graphs(df, single)
            except:
                traceback.print_exc()
    return graphs, '', False, _getCheckSingleOptions(False)

@app.callback(
    Output('dropdown_channels', 'options'),
    Input('dropdown_channels', 'search_value'),
    State("dropdown_channels", "value"),
)
def update_results(regex, value):
    if not regex or len(regex)<3:
        raise PreventUpdate
    options = search_channels(regex)
    #return [o for o in options if o not in value]
    #Must add to options or else selected value will be removed
    if value:
        options = options + [{'label': V, 'value': V} for V in value]
    return options



# Run the app
if __name__ == '__main__':
    ssl_context = None #('cert.pem', 'key_unencrypted.pem')
    app.run_server(host='0.0.0.0', port=8050, debug=True, ssl_context=ssl_context)


import sys
import time

start = time.time()
import os,time,logging,dateutil,json,ssl,urllib,re,numpy,struct,bitshuffle
from pkg_resources import resource_stream
print  (time.time() - start)


start = time.time()
from datahub import *
print  (time.time() - start)



start = time.time()
DEFAULT_SOURCE = os.environ.get("DEFAULT_DATA_SOURCE", "daqbuf")
from datahub.main import run_json
print  (time.time() - start)

# Function to get the top-level modules imported by the application
def get_imported_modules():
    return [name for name, obj in globals().items() if isinstance(obj, type(sys))]

# Print the names of imported modules
imported_modules = get_imported_modules()
print("Imported Modules:", imported_modules)


from datahub import *
from threading import Thread, current_thread
import time
from datahub.utils.reflection import get_meta
import datahub.utils.timing as timing

_logger = logging.getLogger(__name__)

class Source():
    query_index = {}
    instances = set()

    def __init__(self, url=None, backend=None, query_path=None, search_path=None, auto_decompress=False, path=None,
                 known_backends=[], time_type="nano", name=None, **kwargs):
        self.url = url
        if query_path is not None:
            if not url.endswith(query_path):
                self.url = self.url + query_path

        self.search_url = url
        if search_path is not None:
            if not url.endswith(search_path):
                self.search_url = url + search_path

        self.channel_formats = {}
        self.channel_info = {}
        self.last_rec_info = {}
        self.listeners = []
        self.set_backend(backend)
        self.known_backends= known_backends
        if time_type.lower() in ["str", "string"]:
            self.time_type = "str"
        elif time_type.lower() in ["sec", "secs", "seconds"]:
            self.time_type = "sec"
        elif time_type.lower() in ["milli", "millis", "milliseconds"]:
            self.time_type = "milli"
        else:
            self.time_type = "nano"
        self.path = path
        self.query = None
        self.type = type(self).__name__.lower()
        self.query_index= None
        self.query_id = self.user_id = None
        self.processing_thread = None
        self.sprocessing_thread = None
        self.aborted = False
        self.running = False
        self.verbose = False
        self.parallel = False
        self.downsample = False
        self.name = name
        self.auto_decompress = auto_decompress
        self.prefix = ""
        Source.instances.add(self)

    def get_backends(self):
        return self.known_backends
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __str__(self):
        return "%s: %s" % (self.get_desc(), str(self.query))

    def set_backend(self, backend):
        if type(backend) == str:
            if backend.strip().lower() in ["none", "null", ""]:
                backend = None
        self.backend = backend

    def get_backend(self):
        return self.backend

    def get_desc(self):
        return "%s[%s]" % (self.get_id(), (str(self.backend) if self.backend else self.url))

    def set_id(self, id):
        self.user_id = id

    def get_id(self):
        if self.user_id:
            return str(self.user_id)
        return self.query_id

    def get_name(self):
        if self.name:
            return self.name
        return self.get_id()

    def set_path(self, path):
        self.path = path

    def get_path(self):
        if self.path:
            return str(self.path)
        return self.get_id()

    def on_channel_header(self, name, typ, byteOrder, shape, channel_compression, metadata={}):
        if self.prefix:
            name = self.prefix + name
        self.channel_info[name] = [typ, byteOrder, shape, channel_compression, metadata]
        self.last_rec_info[name] = [0.0, 0] #timestamp, index

        for listener in self.listeners:
            try:
                listener.on_channel_header(self, name, typ, byteOrder, shape, None if (self.auto_decompress) else channel_compression, metadata)
            except Exception as e:
                _logger.exception("Error creating channel on listener %s: %s" % (str(listener), str((name, typ, byteOrder, shape, channel_compression))))

    def convert_time(self, timestamp):
        return convert_timestamp(timestamp, self.time_type)

    def adjust_type(self, value):
        if type(value) == int:
            return numpy.int64(value)
        elif type(value) == float:
            return numpy.float64(value)
        elif type(value) == bool:
            return numpy.bool(value)
        elif isinstance(value, list):
            return numpy.array(value)
        return value

    def on_channel_record(self, name, timestamp, pulse_id, value, **kwargs):
        if self.prefix:
            name = self.prefix + name
        if self.downsample:
            now = time.time()
            last_timestamp, last_index = self.last_rec_info[name]
            self.last_rec_info[name][1] = last_index + 1
            # interval is divider
            if self.modulo:
                if last_index % self.modulo != 0:
                    return
            #Downsampling is interval
            if self.interval:
                timespan = now - last_timestamp
                if timespan < self.interval:
                    return
            self.last_rec_info[name][0] = now

        if timestamp is None:
            timestamp = create_timestamp(time.time())

        timestamp = self.convert_time(timestamp)

        if self.auto_decompress:
            [typ, byteOrder, shape, channel_compression, metadata] = self.channel_info[name]
            if channel_compression:
                value = decompress(value, name, channel_compression, shape, typ, byteOrder)

        for listener in self.listeners:
            try:
                listener.on_channel_record(self, name, timestamp, pulse_id, value, **kwargs)
            except Exception as e:
                _logger.exception("Error appending record on listener %s: %s" % (str(listener), str((name, timestamp, pulse_id, value))))

    def on_channel_completed(self, name):
        if self.prefix:
            name = self.prefix + name
        if name in self.channel_info:
            for listener in self.listeners:
                try:
                    listener.on_channel_completed(self, name)
                except Exception as e:
                    _logger.exception("Error completing channel on listener %s: %s" % (str(listener), str(name)))

            try:
                del self.channel_info[name]
            except:
                pass

            try:
                del self.last_rec_info[name]
            except:
                pass


    def on_start(self):
        for listener in self.listeners:
            try:
                listener.on_start(self)
            except Exception as e:
                _logger.exception("Error starting %s: %s" % (str(listener), str(e)))

    def on_stop(self, exception=None):
            for listener in self.listeners:
                try:
                    listener.on_stop(self, exception)
                except Exception as e:
                    _logger.exception("Error stopping %s: %s" % (str(listener), str(e)))

    def req(self, channels, start, end, background=False, **kwargs) -> None:
        query = {
            "channels": channels,
            "start": start,
            "end": end
        }
        query.update(kwargs)
        self.request(query, background)

    def request(self, query, background=False) -> None:
        if self.is_running():
            raise RuntimeError("Ongoing query")
        if self.verbose:
            print (f"Requesting from {self.url}: ", query)

        self.aborted = False
        self.query = query
        self.range = QueryRange(self.query, self)

        self.modulo = self.query.get("modulo", None)
        if type(self.modulo) is str:
            try:
                self.modulo= int (self.modulo)
            except:
                raise RuntimeError("Invalid modulo: ", self.modulo)

        self.interval = self.query.get("interval", None)
        if type(self.interval) is str:
            try:
                self.interval = float(self.interval)
            except:
                raise RuntimeError("Invalid interval: ", self.interval)
        self.downsample = self.interval or self.modulo
        self.query_index = Source.query_index.get(self.type, -1) + 1
        Source.query_index[self.type]=self.query_index
        self.query_id = f"{self.type}_{self.query_index}"

        prefix = self.query.get("prefix", None)
        if prefix:
            has_prefix = str_to_bool(str(prefix))
            if has_prefix == True:
                self.prefix = self.get_name() + ":"
            elif has_prefix == False:
                self.prefix = ""
            else:
                self.prefix = prefix
        else:
            self.prefix = ""

        if background:
            self.processing_thread = Thread(target=self.do_run, args=(query,))
            self.processing_thread.setDaemon(True)
            self.processing_thread.start()
        else:
            self.processing_thread = None
            self.do_run(query)

    def join(self, timeout=None):
        if self.is_thread_running():
            self.processing_thread.join(timeout)

    def is_thread_running(self):
        return self.processing_thread and  self.processing_thread.is_alive()

    def is_running(self):
        return self.running

    def is_aborted(self):
        return self.aborted

    def is_in_processing_thread(self):
        return self.processing_thread == current_thread()


    def abort(self):
        self.aborted = True

    def do_run(self, query):
        self.running = True
        self.on_start()
        try:
            exc = None
            self.run(query)
        except Exception as e:
            exc = e
            raise
        finally:
            self.on_stop(exc)
            self.running = False

    def add_listener(self, listener):
        self.listeners.append(listener)

    def remove_listeners(self):
        self.listeners.clear()

    def close_listeners(self):
        for listener in self.listeners:
            listener.close()
        self.listeners.clear()

    #Utility methods to manage automatically calling on_channel_header on the first stream value
    def receive_channel(self, channel_name, value, timestamp, id, check_changes=False, check_types=False, metadata={}, **kwargs):
        existing = channel_name in self.channel_formats
        if check_types:
            value = self.adjust_type(value);
        if not existing or check_changes:
            try:
                if isinstance(value, str):
                    fmt = typ, shape = "str", None
                else:
                    fmt = typ, shape = (value.dtype, value.shape)
            except:
                _logger.exception("Invalid type of channel %s: %s" % (str(channel_name), str(type(value))))
                return

            if fmt != self.channel_formats.get(channel_name, None):
                if metadata is None:
                    metadata = {}
                metadata["has_id"] = id is not None
                if existing:
                    self.on_channel_completed(channel_name)
                    _logger.warning("Channel %s changed type from %s to %s." % (str(channel_name), str(self.channel_formats.get(channel_name)), str(fmt)))
                    del self.channel_formats[channel_name]
                self.on_channel_header(channel_name, typ, Endianness.LITTLE, shape, None, metadata)
                self.channel_formats[channel_name] = fmt

        self.on_channel_record(channel_name, timestamp, id, value, **kwargs)

    def close_channels(self):
        for channel_name in self.channel_formats.keys():
            self.on_channel_completed(channel_name)
        self.channel_formats = {}
        self.channel_info = {}
        self.last_rec_info = {}

    def close(self):
        if self.is_thread_running():
            if not self.is_in_processing_thread():
                self.abort()
                self.join()
        self.close_channels()
        self.remove_listeners()
        if self in Source.instances:
            Source.instances.remove(self)

    #Virtuals
    def run(self, query):
        pass

    def search(self, regex):
        raise Exception(f"Search not implemented in {self.type}")

    def print_search(self, regex):
        search = self.search(regex)
        if type(search) == str:
            print(search)
        elif search is None:
            print("Empty")
        else:
            print(json.dumps(search, indent=4))

    def get_default_url(self):
        try:
            return getattr(self, "DEFAULT_URL")
        except:
            return None

    def get_default_backend(self):
        try:
            return getattr(self, "DEFAULT_BACKEND")
        except:
            return None

    def time_to_pulse_id(self, tm=time.time()):
        return timing.time_to_pulse_id(tm)

    def pulse_id_to_time(self, id):
        return timing.pulse_id_to_time(id)

    def print_help(self):
        meta = get_meta(self.__class__)
        print(f"Source Name: \n\t{self.type}")
        if (self.__class__.__doc__):
            print("Description:")
            print (f"\t{self.__class__.__doc__.strip()}")
        print(f"Arguments: \n\t[channels {meta}start=None end=None ...]")
        if (self.__class__.__init__.__doc__):
            print (self.__class__.__init__.__doc__)
        default_url = self.get_default_url()
        default_backend = self.get_default_backend()
        if default_url:
            print("Default URL:")
            print(f"\t{default_url}")
        if default_backend:
            print("Default Backend:")
            print(f"\t{default_backend}")
        backends = self.get_backends()
        if backends:
            print("Known Backends:")
            for backend in backends:
                print(f"\t{backend}")

    def _get_pandas(self):
        try:
            import pandas
            return pandas
        except:
            _logger.error("Pandas not installed: cannot report search as dataframe")
            return None

    @staticmethod
    def cleanup():
        for source in list(Source.instances):
            source.close()


import time
import unittest
from datahub import *

backend = "sf-databuffer"
filename = "/Users/gobbo_a/dev/back/daqbuf.h5"
filename = "/Users/gobbo_a/datahub.h5"

channels = ["S10BC01-DBPM010:Q1", "S10BC01-DBPM010:X1"]
start = "2024-02-15T12:41:00Z"
end = "2024-02-15T12:42:00Z"

channels = ["S10BC01-DBPM010:Q1", "S10BC01-DBPM010:X1"]
start = "2024-06-14 09:00:00"
end = "2024-06-14 10:00:00"

query = {
    "channels": channels,
    "start": start,
    "end": end
}

class DataBufferTest(unittest.TestCase):

    def test_dataframe(self):
        with Daqbuf(backend=backend, cbor=True, parallel=True) as source:
            table = Table()
            source.add_listener(table)
            source.request(query)
            dataframe_cbor = table.as_dataframe(Table.PULSE_ID)
            dataframe_cbor.reindex(sorted(dataframe_cbor.columns), axis=1)
            print(dataframe_cbor)

    def test_listeners(self):
        parallel = True
        for index, parallel in [(Table.PULSE_ID, False), (Table.TIMESTAMP, True)]:
            with Daqbuf(backend=backend, cbor=True, parallel=parallel) as source:
                table = Table()
                source.add_listener(table)
                source.request(query)
                dataframe_cbor = table.as_dataframe(index)
                dataframe_cbor.reindex(sorted(dataframe_cbor.columns), axis=1)
                print (dataframe_cbor)

            with Daqbuf(backend=backend, cbor=False, parallel=parallel) as source:
                table = Table()
                source.add_listener(table)
                source.request(query)
                dataframe_json = table.as_dataframe(index)
                dataframe_json.reindex(sorted(dataframe_json.columns), axis=1)
                print(dataframe_json)
            self.assertEqual(dataframe_cbor.equals(dataframe_json), True)

    def test_save(self):
        s = time.time()
        with Daqbuf(backend=backend, cbor=True, parallel=True) as source:
            hdf5 = HDF5Writer(filename)
            source.add_listener(hdf5)
            source.req(channels, start, end)
        print (time.time()-s)


    def test_listener(self):
        last = None
        class Listener(Consumer):
            def on_start(self, source):
                pass

            def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
                print(f"Started: {name}")

            def on_channel_record(self, source, name, timestamp, pulse_id, value):
                #print(f"{timestamp} {name}={str(value)} ")
                nonlocal last
                last = timestamp, pulse_id, value

            def on_channel_completed(self, source, name):
                timestamp, pulse_id, value = last
                timestr = convert_timestamp(timestamp, "str")
                print(f"Completed: {name}: {last} - {timestr}")

            def on_stop(self, source, exception):
                pass


        s = time.time()
        with Daqbuf(backend=backend, cbor=True, parallel=True) as source:
            source.add_listener(Listener())
            source.request(query)
        print (time.time()-s)


    def test_waveform(self):

        with Daqbuf(backend=backend, cbor=True, parallel=True) as source:
            stdout = Stdout()
            source.add_listener(stdout)
            source.req(["SARFE10-PSSS059:SPECTRUM_X"], "2024-05-07 16:00:00", "2024-05-07 16:00:01", False)


    def test_backends(self):
        with Daqbuf() as source:
            print (source.get_backends())
            print(source.get_backends())

    def test_sf_archiver(self):

        with Daqbuf(backend="sf-archiver", cbor=True, parallel=True) as source:
            stdout = Stdout()
            source.add_listener(stdout)
            source.req(["S10CB05-RBOC-DCP10:REF-POWER-AVG"], -1000, 0, False)


    def test_binned(self):
        query["bins"]=100
        with Daqbuf(backend=backend, cbor=True, parallel=False) as source:
            table = Table()
            source.add_listener(table)
            source.request(query)
            df = table.as_dataframe(Table.TIMESTAMP)
            df.reindex(sorted(df.columns), axis=1)
            print(df)

    def test_binned2(self):
        query["bins"]=100
        with Daqbuf(backend=backend, cbor=True, parallel=False) as source:
            stout = Stdout()
            source.add_listener(stout)
            source.request(query)

    def test_binned3(self):
        query["bins"]=100
        with Daqbuf(backend=backend, cbor=True, parallel=False) as source:
            plot = Plot()
            source.add_listener(plot)
            source.request(query)

    def test_binned4(self):
        query["bins"]=100
        with Daqbuf(backend=backend, cbor=True, parallel=False, time_type="millis") as source:
            stdout = Stdout()
            hdf5 = HDF5Writer(filename)
            source.add_listener(hdf5)
            source.add_listener(stdout)
            source.request(query)

if __name__ == '__main__':
    unittest.main()



import unittest
from datahub import *
import time

backend = "sf-databuffer"
filename = "/Users/gobbo_a/dev/back/databuffer.h5"


channels = ["S10BC01-DBPM010:Q1", "S10BC01-DBPM010:Y2"]
start = "2024-01-05T08:50:00.000Z"
end = "2024-01-05T08:50:05.000Z"

channels = ["S10BC01-DBPM010:Q1", ]
channels = ["SARFE10-PSSS059:SPECTRUM_X"]
start = "2024-02-14 08:50:00.000"
end = "2024-02-14 08:50:05.000"

#channels = ["SARES20-CAMS142-M5.processing_parameters"]
#start = "2024-02-14 16:50:00.000"
#end = "2024-02-14 16:50:05.000"


#start = 20319949320
#end =   20319949818

#start = 200
#end =   500

#start = -15.0
#end =   -12.0

query = {
    "channels": channels,
    "start": start,
    "end": end
}

class DataBufferTest(unittest.TestCase):
    def setUp(self):
        self.source = DataBuffer(backend=backend)

    def tearDown(self):
        self.source.close()

    def test_listeners(self):
        hdf5 = HDF5Writer(filename)
        stdout = Stdout()
        table = Table()
        self.source.add_listener(hdf5)
        self.source.add_listener(stdout)
        self.source.add_listener(table)
        self.source.request(query)
        dataframe = table.as_dataframe()
        self.source.close_listeners()


if __name__ == '__main__':
    unittest.main()



import time
import unittest
from datahub import *

class Listener (Consumer):
    def on_start(self, source):
        pass

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        print(f"Started: {name}")

    def on_channel_record(self, source, name, timestamp, pulse_id, value):
        print(f"{timestamp} {name}={str(value)} ")

    def on_channel_completed(self, source, name):
        print(f"Completed: {name}")

    def on_stop(self, source, exception):
        pass


class BsreadTest(unittest.TestCase):

    def test_listeners(self):
        with Bsread(url="tcp://localhost:9999", mode="PULL", time_type="str") as source:
            listener = Listener()
            source.add_listener(listener)
            source.req(["UInt8Scalar", "Float32Scalar"], 0.0, 0.2)

if __name__ == '__main__':
    unittest.main()



import sys
import time

from datahub import is_null_str
from datahub.utils.timing import *

_logger = logging.getLogger(__name__)

class QueryRange():
    """
    The query range is defined by the parameters "start" and "end" in the parameters.
    If  type is float or float-convertible string, it means the time in seconds since the epoch, or
        if the value is lesser than max_relative_time, an offest to current epoch.
    If type is an int or int-convertible string, it means an id or
        if the value is lesser than max_relative_is, an offset to current id`.
    If type is str it is a date in ISO format.

    Optionally parameters "start_id", end_id", "start_tm" and end_tm" can be used to avoid type errors,
    """
    RANGE_DEFAULTS_PULSE_ID = -(5 * 365 * 24 * 3600), (365 * 24 * 3600)  #From 5 years ago to one year from now

    def __init__(self, query, source=None):
        now = time.time()
        self.start = self._check_str(query.get("start", None))
        self.end = self._check_str(query.get("end", None))
        self.source = source
        range_defaults_pulse_id = self.time_to_id(now + QueryRange.RANGE_DEFAULTS_PULSE_ID[0]), \
                                  self.time_to_id(now + QueryRange.RANGE_DEFAULTS_PULSE_ID[1])
        if type(self.start) == float: #timestamp
            if range_defaults_pulse_id[0] < self.start <range_defaults_pulse_id[1]:
                self.start = int(self.start) #assumes pulse id
        if type(self.end) == float: #timestamp
            if range_defaults_pulse_id[0] < self.end <range_defaults_pulse_id[1]:
                self.end = int(self.end) #assumes pulse id

        start_id = query.get("start_id", None)
        if start_id is not None:
            self.start = int(start_id)
        end_id = query.get("end_id", None)
        if end_id is not None:
            self.end = int(end_id)

        start_tm = self._check_str(query.get("start_tm", None))
        if start_tm is not None:
            self.start = float(start_tm)
        end_tm = self._check_str(query.get("end_tm", None))
        if end_tm is not None:
            self.end = end_tm if (type(end_tm) == str) else float(end_tm)

        if self.start is None:
            if type(self.end) == int:
                self.start = 0
            elif type(self.end) == str:
                self.start = self.seconds_to_string(now)
            else:
                self.start = 0.0
        if self.end is None:
            if type(self.start) == int:
                self.end = 0
            elif type(self.start) == str:
                self.end = self.seconds_to_string(now)
            else:
                self.end = 0.0
        #import pytz
        #self.utc_tz = pytz.utc
        #self.local_tz = pytz.timezone('Europe/Zurich')
        self.max_relative_time = 1000000.0 #in secs ~=11 days
        self.max_relative_id = 100000000

        #Seconds
        self.start_sec, self.start_str, self.start_id, self.start_type = self._parse_par(self.start, now, True)
        self.end_sec, self.end_str, self.end_id, self.end_type = self._parse_par(self.end, now, False)

        if self.start_sec > self.end_sec:
            raise Exception("Invalid query range: %s to %s" % (self.start, self.end))

    def _check_str(self, par):
        if type(par) == str:
            try:
                return float(par)
            except:
                try:
                    return int(par)
                except:
                    if is_null_str(par):
                        return None
                    return par
        return par

    def _parse_par(self, par, now, start):
        if par is None: #Absent parameters mean the current time
            par = 0.0
        if type(par) == float:
            sec = par
            if par <= self.max_relative_time:
                sec = sec + now
            st = self.seconds_to_string(sec)
            id = None
            typ = "time"
        #Date
        elif type(par) == str:
            sec = self.string_to_seconds(par)
            st = par
            id = None
            typ = "date"
        #ID
        elif type(par) == int:
            id = par
            if par <= self.max_relative_id:
                id = id + self.time_to_id()
            sec = self.id_to_time(id)
            offset = PULSE_ID_INTERVAL/2
            sec = sec + (-offset if start else offset)
            st = self.seconds_to_string(sec)
            typ = "id"
        else:
            raise Exception("Invalid parameter value: " + str(par))
        return sec, st, id, typ

    def time_to_id(self, tm=time.time()):
        if self.source is not None:
            return self.source.time_to_pulse_id(tm)
        return time_to_pulse_id(tm)

    def id_to_time(self, id):
        if self.source is not None:
            return self.source.pulse_id_to_time(id)
        return pulse_id_to_time(id)

    def wait_time(self, target_time):
        current_time = time.time()
        while current_time < target_time:
            time.sleep(target_time - current_time)
            current_time = time.time()

    def wait_start(self, delay=0.0):
        self.wait_time(self.get_start_sec() + delay)

    def wait_end(self, delay=0.0):
        self.wait_time(self.get_end_sec() + delay)

    def has_started(self, tm=None):
        if tm is None:
            tm = time.time()
        return tm >= self.get_start_sec()

    def has_ended(self, tm=None):
        if tm is None:
            tm = time.time()
        return tm > self.get_end_sec()

    def is_running(self, tm=None):
        if tm is None:
            tm = time.time()
        return self.has_started(tm) and not self.has_ended(tm)

    def get_start(self):
        return self.start

    def get_end(self):
        return self.end

    def get_start_sec(self):
        return self.start_sec

    def get_end_sec(self):
        return self.end_sec

    def get_start_str(self):
        return self.start_str

    def get_end_str(self):
        return self.end_str

    def get_start_str_iso(self):
        start = self.get_start_str()
        start = self.string_to_datetime(start)
        return datetime.isoformat(start)

    def get_end_str_iso(self):
        end = self.get_end_str()
        end = self.string_to_datetime(end)
        return datetime.isoformat(end)

    def get_start_id(self):
        return self.start_id

    def get_end_id(self):
        return self.end_id

    def get_start_type(self):
        return self.start_type

    def get_end_type(self):
        return self.end_type

    def get_type(self):
        return self.get_end_type()

    def is_start_by_id(self):
        return self.start_id is not None

    def is_end_by_id(self):
        return self.end_id is not None

    def seconds_to_string(self, seconds, utc=True):
        return timestamp_to_string(seconds, utc)

    def string_to_seconds(self, date_string):
        return string_to_timestamp(date_string)

    def string_to_datetime(self, date_string):
        if isinstance(date_string, str):
            date = string_to_datetime(date_string)
        elif isinstance(date_string, datetime.datetime):
            date = date_string
        else:
            raise ValueError("Unsupported date type: " + type(date_string))
        if date.tzinfo is None:  # localize time if necessary
            try:
                import pytz
                date = pytz.timezone('Europe/Zurich').localize(date)
            except Exception as ex:
                _logger.error(ex)
        return date

    def __str__(self):
       return f"Range from {self.get_start_str()} ({self.get_start_id()}) to {self.get_end_str()} ({self.get_end_id()})"




from pkg_resources import resource_stream
import os

def version():
    with resource_stream(__name__, "package_version.txt") as res:
        return res.read()[:-1].decode()

def str_to_bool(str):
    if str.lower() in ('true', 'yes', '1'):
        return True
    if str.lower() in ('false', 'no', '0'):
        return False
    return None

def is_null_str(str):
    return (str is None) or (str.lower() == "null") or (str.lower() == "none")

class ProtocolError(RuntimeError):
    def __init__(self):
        super().__init__("ProtocolError")

from datahub.utils.timing import *
from datahub.utils.net import *
from datahub.utils.compression import *
from datahub.utils.range import QueryRange
from datahub.source import Source
from datahub.sources.retrieval import Retrieval
from datahub.sources.bsread import Bsread, BsreadStream
from datahub.sources.array10 import Array10
from datahub.sources.epics import Epics
from datahub.sources.camserver import Pipeline
from datahub.sources.camserver import Camera
from datahub.sources.databuffer import DataBuffer
from datahub.sources.dispatcher import Dispatcher
from datahub.sources.daqbuf import Daqbuf
from datahub.sources.stddaq import Stddaq
from datahub.sources.redis import Redis, RedisStream
from datahub.consumer import Consumer
from datahub.consumers.h5 import HDF5Writer
from datahub.consumers.txt import TextWriter
from datahub.consumers.stdout import Stdout
from datahub.consumers.table import Table
from datahub.consumers.pshell import PShell
from datahub.consumers.plot import Plot
from datahub.consumers.merger import Merger

def cleanup():
    Source.cleanup()
    Consumer.cleanup()

KNOWN_CONSUMERS = {
    "hdf5": ("f", HDF5Writer),
    "txt":  ("x", TextWriter),
    "print": ("p", Stdout),
    "plot": ("m", Plot),
    "pshell": ("ps", PShell)
}

KNOWN_SOURCES = {
    "epics": Epics,
    "bsread": Bsread,
    "pipeline": Pipeline,
    "camera": Camera,
    "databuffer": DataBuffer,
    "retrieval": Retrieval,
    "dispatcher": Dispatcher,
    "daqbuf": Daqbuf,
    "array10": Array10,
    "redis": Redis,
    "stddaq": Stddaq
    }

DEFAULT_SOURCE = os.environ.get("DEFAULT_DATA_SOURCE", "daqbuf")
#from datahub.main import run_json



import unittest
from datahub import *

url = "localhost:54321"
filename = "/Users/gobbo_a/dev/back/epics.h5"
channels = ["TESTIOC:TESTCALCOUT:Input", "TESTIOC:TESTSINUS:SinCalc", "TESTIOC:TESTWF2:MyWF"]
start = None
end = 3.0
query = {
    "channels": channels,
    "start": start,
    "end": end
}



class EpicsTest(unittest.TestCase):
    def setUp(self):
        self.source = Epics()

    def tearDown(self):
        cleanup()


    def test_listeners(self):
        hdf5 = HDF5Writer(filename)
        stdout = Stdout()
        table = Table()
        #self.source.set_id("bsread")
        self.source.add_listener(hdf5)
        self.source.add_listener(stdout)
        self.source.add_listener(table)
        self.source.request(query)
        dataframe = table.as_dataframe()
        print (dataframe.columns)
        if channels:
            self.assertEqual(list(dataframe.keys()), channels)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# HTTP Utilities
########################################################################################################################

import datahub
import json
from http.client import HTTPSConnection, HTTPConnection
import ssl
import urllib
import re
import sys
import logging

_logger = logging.getLogger(__name__)

def get_host_port_from_stream_address(stream_address):
    if stream_address.startswith("ipc"):
        return stream_address.split("//")[1], -1
    source_host, source_port = stream_address.rsplit(":", maxsplit=1)
    if "//" in source_host:
        source_host = source_host.split("//")[1]
    return source_host, int(source_port)

def create_http_conn(up):
    if type(up) == str:
        up = urllib.parse.urlparse(up)
    if up.scheme == "https":
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        ctx.check_hostname = False
        port = up.port
        if port is None:
            port = 443
        conn = HTTPSConnection(up.hostname, port, context=ctx)
    else:
        port = up.port
        if port is None:
            port = 80
        conn = HTTPConnection(up.hostname, port)
    return conn


def http_req(method, url, conn=None):
    headers = {
        "X-PythonDataAPIPackageVersion": datahub.version(),
        "X-PythonDataAPIModule": __name__,
        "X-PythonVersion": re.sub(r"[\t\n]", " ", str(sys.version)),
        "X-PythonVersionInfo": str(sys.version_info),
    }
    up = urllib.parse.urlparse(url)
    if conn is None:
        conn = create_http_conn(up)
    conn.request(method, up.path, None, headers)
    #return conn.getresponse()
    return conn


def http_data_query(query, url, method = "POST", content_type="application/json", accept="application/octet-stream", conn=None):
    headers = {
        "Content-Type": content_type,
        "Accept": accept,
        "X-PythonDataAPIPackageVersion": datahub.version(),
        "X-PythonDataAPIModule": __name__,
        "X-PythonVersion": re.sub(r"[\t\n]", " ", str(sys.version)),
        "X-PythonVersionInfo": str(sys.version_info),
    }
    up = urllib.parse.urlparse(url)
    if conn is None:
        conn = create_http_conn(up)
    if method == "GET":
        params = urllib.parse.urlencode(query)
        url = f'{url}?{params}'
        conn.request("GET", url, headers=headers)
    else:
        body = json.dumps(query)
        conn.request(method, up.path, body, headers)
    return conn

def get_json(url):
    conn = http_req("GET", url)
    try:
        res = conn.getresponse()
        body = res.read().decode()
        try:
            return json.loads(body)
        except:
            _logger.error(f"can not parse request status as json\n" + body)
            return body
    finally:
        conn.close()

def save_raw(query, url, fname):
    conn = http_data_query(query, url)
    try:
        s = conn.getresponse()
        with open(fname, "wb") as f1:
            while True:
                buf = s.read()
                if buf is None:
                    break
                if len(buf) < 0:
                    raise RuntimeError()
                if len(buf) == 0:
                    break
                f1.write(buf)
    finally:
        conn.close()




from datahub import *
import io
from http.client import IncompleteRead

_logger = logging.getLogger(__name__)

KNOWN_BACKENDS = ["sf-databuffer", "sf-imagebuffer", "hipa-archive"]

class ProcessChannelHeaderResult:
    def __init__(self):
        self.error = False
        self.empty = False
        self.channel_info = None
        self.channel_name = None
        self.extractor_writer = None
        self.compression = None
        self.shape = None

def extractor_raw_data_blob(ts, pulse, buf, name, data_type, shape):
    return ts, pulse, buf

def extractor_basic_scalar(ts, pulse, buf, name, data_type, shape):
    value = numpy.frombuffer(buf, dtype=data_type)[0]
    return ts, pulse, value

def extractor_basic_shaped(ts, pulse, buf, name, data_type, shape):
    value = numpy.reshape(numpy.frombuffer(buf, dtype=data_type), shape)
    return ts, pulse, value

def resolve_struct_dtype(data_type: str, byte_order: str) -> str:
    if data_type is None:
        None
    data_type = data_type.lower()
    if data_type == "float64":
        dtype = 'd'
    elif data_type == "uint8":
        dtype = 'B'
    elif data_type == "int8":
        dtype = 'b'
    elif data_type == "uint16":
        dtype = 'H'
    elif data_type == "int16":
        dtype = 'h'
    elif data_type == "uint32":
        dtype = 'I'
    elif data_type == "int32":
        dtype = 'i'
    elif data_type == "uint64":
        dtype = 'Q'
    elif data_type == "int64":
        dtype = 'q'
    elif data_type == "float32":
        dtype = 'f'
    elif data_type == "bool8":
        dtype = '?'
    elif data_type == "bool":
        dtype = '?'
    elif data_type == "character":
        dtype = 'c'
    elif data_type == "string":
        dtype = "str"
    else:
        raise RuntimeError(f"unsupported dta type {data_type} {byte_order}")
    if byte_order is not None:
        x = byte_order.upper()
        if x not in ["LITTLE_ENDIAN", "BIG_ENDIAN"]:
            raise RuntimeError(f"unexpected byte order {byte_order}")
    if dtype != "str":
        if byte_order is not None and byte_order.upper() == "BIG_ENDIAN":
            dtype = ">" + dtype
        else:
            dtype = "<" + dtype
    return dtype



def process_channel_header(msg):
    name = msg["name"]
    ty = msg.get("type")
    # If no data could be found for this channel, then there is no `type` key and we stop here:
    if ty is None:
        res = ProcessChannelHeaderResult()
        res.empty = True
        res.channel_name = name
        return res
    dtype = resolve_struct_dtype(ty, msg.get("byteOrder"))
    if dtype is None:
        raise RuntimeError("unsupported dtype {} for channel {}".format(dtype, name))
    shape = list(reversed(msg.get("shape", [])))

    compression = msg.get("compression")
    # Older data api services indicate no-compression as `0` or even `"0"`
    # we handle these cases here
    if compression is not None:
        compression = int(compression)
    if compression == 0:
        compression = None
    if compression is None:
        if shape == [1]:
            # NOTE legacy compatibility: historically a shape [1] is treated as scalar
            # Which channels actually rely on this?
            _logger.warning(f"Received scalar-like shape, convert to true scalar  {name}")
            shape = []
        if len(shape) == 0:
            if dtype == "str":
                #extractor_writer = None
                extractor_writer = lambda ts, pulse, b: extractor_raw_data_blob(ts, pulse, b, name, data_type, shape)
            else:
                data_type = numpy.dtype(msg.get("type")).newbyteorder('<' if msg.get("byteOrder") == "LITTLE_ENDIAN" else ">")
                extractor_writer = lambda ts, pulse, b: extractor_basic_scalar(ts, pulse, b, name, data_type, shape)
        elif len(shape) > 0:
            if dtype == "str":
                raise RuntimeError("not yet supported, please report a channel that uses arrays of strings.")
            else:
                data_type = numpy.dtype(msg.get("type")).newbyteorder('<' if msg.get("byteOrder") == "LITTLE_ENDIAN" else ">")
                extractor_writer = lambda ts, pulse, b: extractor_basic_shaped(ts, pulse, b, name, data_type, shape)
        else:
            raise RuntimeError("unexpected  shape {shape}  channel {name}")
    elif compression == 1: #BITSHUFFLE_LZ4
        if dtype == "str":
            if len(shape) == 0:
                extractor_writer = lambda ts, pulse, b: extractor_raw_data_blob(ts, pulse, b, name, data_type, shape)
            else:
                raise RuntimeError("Arrays of strings not supported")
        else:
            if len(shape) == 0:
                raise RuntimeError(f"Compression not supported on scalar numeric data {name}  shape {shape}  dtype {dtype}")
            else:
                data_type = numpy.dtype(msg.get("type")).newbyteorder('<' if msg.get("byteOrder") == "LITTLE_ENDIAN" else ">")
                extractor_writer = lambda ts, pulse, b: extractor_raw_data_blob(ts, pulse, b, name, data_type, shape)
    else:
        raise RuntimeError(f"Compression type {compression} is not supported")

    res = ProcessChannelHeaderResult()
    res.channel_info = msg
    res.channel_name = name
    res.extractor_writer = extractor_writer
    res.compression = Compression.BITSHUFFLE_LZ4 if compression else None
    res.shape = shape
    return res

class Retrieval(Source):
    """
    Retrieves data from the old Retrieval.
    """

    DEFAULT_URL = os.environ.get("RETRIEVAL_DEFAULT_URL", "https://data-api.psi.ch/api/1")
    DEFAULT_BACKEND = os.environ.get("RETRIEVAL_DEFAULT_BACKEND", "sf-databuffer")

    def __init__(self, url=DEFAULT_URL, backend=DEFAULT_BACKEND, path=None, delay=1.0, **kwargs):
        """
        url (str, optional): Retrieval URL. Default value can be set by the env var RETRIEVAL_DEFAULT_URL.
        backend (str, optional): Retrieval backend. Default value can be set by the env var RETRIEVAL_DEFAULT_BACKEND.
        path (str, optional): hint for the source location in storage or displaying.
        delay (float, optional): Wait time for channels to be uploaded to storage before retrieval.
        """
        if url is None:
            raise RuntimeError("Invalid URL")

        if backend is not None:
            if backend not in KNOWN_BACKENDS:
                _logger.warning("Unknown backend: " + str(backend))

        Source.__init__(self, url=url, backend=backend, query_path="/query", search_path="/channels", path=path,
                        known_backends=KNOWN_BACKENDS, **kwargs)
        self.delay = delay

    def _get_range(self, start_expansion=False, end_expansion=False):
        #TODO: Only searching by date?
        start = self.range.get_start_str_iso()
        end = self.range.get_end_str_iso()
        return {
            "type": "date",
            "startDate": start,
            "endDate": end
        }

    def _get_request_status(self, url, reqid):
        url_status = re.sub(r"/[^/]+$", "/requestStatus/" + reqid, url)
        return get_json(url_status)

    def _get_request_status_from_immediate_error(self, url, response):
        response_body = response.read(1024).decode()
        try:
            err = json.loads(response_body)
            reqid = err["requestId"]
            url_status = re.sub(r"/[^/]+$", "/requestStatus/" + reqid, url)
            err = get_json(url_status)
            if err: _logger.error(err)
            return err
        except:
            _logger.error(f"can not parse error message as json:\n{response_body}")
            raise

    def run(self, query):
        self.range.wait_end(delay=1.0)

        json = {}
        json["channels"] = query["channels"]
        json["range"] = self._get_range()
        if self.backend is not None:
            json["defaultBackend"] = self.backend
        conn = http_data_query(json, self.url)
        try:
            response = conn.getresponse()
            if response.status != 200:
                _logger.error(f"Unable to retrieve data: {response.status}")
                status = self._get_request_status_from_immediate_error(self.url, response)
                raise RuntimeError(f"Unable to retrieve data  {str(status)}")
            try:
                self.read(io.BufferedReader(response))
                reqid = response.headers["x-daqbuffer-request-id"]
                stat = self._get_request_status(self.url, reqid)
                if stat.get("errors") is not None:
                    raise RuntimeError("request error")
            except (ProtocolError, RuntimeError) as e:
                _logger.error(f"error during request  {e}")
                reqid = response.headers["x-daqbuffer-request-id"]
                stat = self._get_request_status(self.url, reqid)
                _logger.error(f"request status: {stat}")
                raise
        finally:
            conn.close()


    def read(self, stream):
        try:
            return self.read_throwing(stream)
        except IncompleteRead:
            _logger.error("unexpected end of input")
            raise ProtocolError()

    def read_throwing(self, stream):
        current_channel_info = None
        current_channel_name = None
        header = None

        while True:
            bytes_read = stream.read(4)
            if len(bytes_read) != 4:
                if current_channel_name is not None:
                    self.on_channel_completed(current_channel_name)
                break
            length = struct.unpack('>i', bytes_read)[0]
            bytes_read = stream.read(length)
            if len(bytes_read) != length:
                raise RuntimeError("unexpected EOF")
            mtype = struct.unpack('b', bytes_read[:1])[0]

            if mtype == 1 and (current_channel_info is not None):
                timestamp = struct.unpack('>q', bytes_read[1:9])[0]
                pulse_id = struct.unpack('>q', bytes_read[9:17])[0]
                raw_data_blob = bytes_read[17:]
                timestamp, pulse_id, value = header.extractor_writer(timestamp, pulse_id, raw_data_blob)
                self.on_channel_record(current_channel_name, timestamp, pulse_id, value)

            # Channel header message
            # A json message that specifies among others data type, shape, compression flags.
            elif mtype == 0:
                if current_channel_name is not None:
                    self.on_channel_completed(current_channel_name)
                current_channel_name = None
                current_channel_info = None
                try:
                    msg = json.loads(bytes_read[1:])
                    res = process_channel_header(msg)
                except Exception as e:
                    raise RuntimeError("Can not process channel header") from e
                if res.error:
                    _logger.error(f"Can not parse channel header message: {msg}")
                elif res.empty:
                    _logger.debug(f"No data for channel {res.channel_name}")
                else:
                    if "type" not in msg:
                        raise RuntimeError()
                    header = res
                    current_channel_info = res.channel_info
                    current_channel_name = res.channel_name
                    if current_channel_info.get('shape', None) is not None:
                         current_channel_info['shape'].reverse()
                    self.on_channel_header(current_channel_name, current_channel_info['type'],
                                                    current_channel_info['byteOrder'],
                                                    current_channel_info['shape'],
                                                    res.compression)
            bytes_read = stream.read(4)
            length_check = struct.unpack('>i', bytes_read)[0]
            if length_check != length:
                raise RuntimeError(f"corrupted file reading {length} {length_check}")

    def search(self, regex):
        import requests
        res = requests.get(self.search_url, {"regex": regex})
        res.raise_for_status()
        ret = res.json()
        return ret




from datahub import *
from datahub.utils.data import *
import redis
import time
import threading
import random

DEFAULT_URL = os.environ.get("REDIS_DEFAULT_URL", 'sf-daqsync-18:6379')
HOST, PORT = get_host_port_from_stream_address(DEFAULT_URL)

def ingest(r, stream, data, max_stream_lenght=1000):
    if not isinstance(stream, (list, tuple)):
        stream, data = [stream], [data]
    if len(stream) == 1:
        r.xadd(stream[0], data[0], maxlen=max_stream_lenght)
    else:
        pipeline = r.pipeline()
        for i in range(len(stream)):
            pipeline.xadd(stream[i], data[i], maxlen=max_stream_lenght)
        pipeline.execute()

def ingest_bsdata(r, pulse_id, timestamp, data, max_stream_lenght=1000):
    channels = list(data.keys())
    data = [{'channel': channel, 'timestamp': timestamp, 'value': encode(data[channel]), 'id': pulse_id} for channel in channels]
    ingest(r, channels, data, max_stream_lenght)


def wait_new_id(id):
    while True:
        time.sleep(0.001)  # 100Hz
        now = time.time()
        new_id = time_to_pulse_id(now)
        if new_id != id:
            return new_id, create_timestamp(now)

def produce_single(channel):
    try:
        with redis.Redis(host=HOST, port=PORT, db=0) as r:
            now = time.time()
            id, timestamp = time_to_pulse_id(now), create_timestamp(now)
            while(True):
                value = random.random()
                #data = {'channel': channel, 'timestamp': timestamp, 'value': encode(value), 'id': id}
                #ingest(r, channel, data)
                data = {channel: value}
                ingest_bsdata(r, id, timestamp, data)
                id, timestamp = wait_new_id(id)

    except Exception as e:
        print(e)

def produce_batch(channels):
    try:
        with redis.Redis(host=HOST, port=PORT, db=0) as r:
            now = time.time()
            id, timestamp = time_to_pulse_id(now), create_timestamp(now)
            while(True):
                #data =[{'channel': channel, 'timestamp': timestamp, 'value': encode(random.random()), 'id': id} for channel in channels]
                #ingest(r, channels, data)
                data = {channel: random.random() for channel in channels}
                ingest_bsdata(r, id, timestamp, data)
                id, timestamp = wait_new_id(id)
    except Exception as e:
        print(e)


if __name__ == '__main__':
    batch = True
    channels = range(1, 4)
    threads = []
    if batch:
        thread = threading.Thread(target=produce_batch, args=([f"channel{i}" for i in channels],))
        threads.append(thread)
        thread.start()
    else:
        for channel in channels:
            thread = threading.Thread(target=produce_single, args=(f"channel{channel}",))
            threads.append(thread)
            thread.start()
    for thread in threads:
        thread.join()



import logging
import numpy
import os
from datahub import Consumer

_logger = logging.getLogger(__name__)

class TextWriter(Consumer):

    def __init__(self, folder: str, **kwargs):
        Consumer.__init__(self, **kwargs)
        self.folder = folder
        self.files = {}

    def on_start(self, source):
        os.makedirs(self.folder, exist_ok=True)
        self.files[source]={}

    def on_stop(self, source, exception):
        try:
            for f in self.files[source]:
                f.close()
            del self.files[source]
        except:
            pass

    def get_path(self, source, name):
        return f"{self.folder}/{name}"

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        if len(shape)<2:
            filename = self.get_path(source, name)
            self.files[source][name] = open(filename, 'a')

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        file = self.files[source].get(name, None)
        if file:
            trailer = ""
            if type(value) == bytes:
                value = f"bytes({len(value)})"
            elif isinstance(value, numpy.ndarray) and (len(value.shape) == 1):
                #value = numpy.array_str(value, max_line_width=numpy.inf)
                value = ' '.join(map(str, value))
                trailer = " " #Marker for array
            else:
                value = str(value)
            file.write(f"{str(timestamp)}\t{str(pulse_id)}\t{value}{trailer}\n")

    def on_channel_completed(self, source, name):
        try:
            del self.files[source][name]
        except:
            pass

    def on_close(self):
        for src, files in self.files.items():
            try:
                for f in files:
                    f.close()
            except:
                pass
        self.files = {}



import logging
import time
from datahub.utils.data import *
from datahub.utils.checker import check_msg
from datahub.utils.timing import get_utc_offset, string_to_timestamp, time_to_pulse_id

_logger = logging.getLogger(__name__)

class Align():
    def __init__(self, callback,  channels=None, range=None, filter=None, partial_msg=True, size_buffer=1000, utc_timestamp=False):
        self.set_channels(channels)
        self.size_buffer = size_buffer
        self.aligned_data = MaxLenDict(maxlen=int(size_buffer*1.2))
        self.partial_msg=partial_msg
        self.on_msg=callback
        self.range = range
        self.filter = filter
        self.sent_id = -1
        self.utc_offset = get_utc_offset() if utc_timestamp else 0

    def set_channels(self, channels):
        self.channels = channels
        self.no_channels = 0 if channels==None else len(channels)

    def add(self, id, timestamp, channel, value):
        if not id:
            id = time_to_pulse_id(timestamp)
        if id not in self.aligned_data:
            self.aligned_data[id] = {"timestamp": timestamp}
        self.aligned_data[id][channel] = value

    def reset(self):
        self.aligned_data.clear()

    def set_range(self, range):
        self.range = range

    def set_filter(self, filter):
        self.filter = filter

    def process(self):
        keys_in_order = sorted(list(self.aligned_data.keys()))
        last_complete_id = -1
        for id in [keys_in_order[i] for i in range(len(keys_in_order) - 1, 0, -1)]:
            if len(self.aligned_data[id]) == (self.no_channels + 1):
                last_complete_id = id
                break

        for i in range(len(keys_in_order)):
            id = keys_in_order[i]
            complete = len(self.aligned_data[id]) == (self.no_channels + 1)
            done = complete or (last_complete_id > id) or (len(self.aligned_data) > self.size_buffer)
            if not done:
                break
            msg = self.aligned_data.pop(id)
            if complete or self.partial_msg:
                if self.sent_id >= id:
                    _logger.warning(f"Invalid ID {id} - last sent ID {self.sent_id}")
                else:
                    timestamp = msg.pop("timestamp", None)
                    started = True
                    if self.range:
                        if isinstance(timestamp, str):
                            timestamp_s = string_to_timestamp(timestamp)
                        else:
                            timestamp_s = timestamp/1e9 if isinstance(timestamp, int) else timestamp
                        timestamp_s = timestamp_s + self.utc_offset if timestamp_s else time.time()
                        started = self.range.has_started(timestamp_s)
                    if started:
                        try:
                            if not self.filter or self.is_valid(self.filter, id, timestamp, msg):
                                self.on_msg(id, timestamp, msg)
                        except Exception as e:
                            _logger.exception("Error receiving data: %s " % str(e))
                    self.sent_id = id
            else:
                logging.debug(f"Discarding partial message: {id}")

    def is_valid(self, filter, id, timestamp, msg):
        try:
            return check_msg(msg, filter)
        except Exception as e:
            _logger.warning("Error processing filter: %s " % str(e))
            return False


import inspect

class Consumer:
    instances = set()

    def __init__(self, **kwargs):
        Consumer.instances.add(self)

    def on_start(self, source):
        pass

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        pass

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        pass

    def on_channel_completed(self, source, name):
        pass

    def on_stop(self, source, exception):
        pass

    def on_close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


    def close(self):
        self.on_close()
        if self in Consumer.instances:
            Consumer.instances.remove(self)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    @staticmethod
    def cleanup():
        for consumer in list(Consumer.instances):
            consumer.close()



import zmq

from datahub import *

GENERATE_ID = False
_logger = logging.getLogger(__name__)

class Array10(Source):
    """
    Retrieves data from an Array10 stream.
    """
    DEFAULT_URL = os.environ.get("ARRAY10_DEFAULT_URL", None)
    def __init__(self, url=DEFAULT_URL, mode="SUB", path=None, reshape=False, **kwargs):
        """
        url (str, optional): Stream URL. Default value can be set by the env var ARRAY10_DEFAULT_URL.
        mode (str, optional): "SUB" or "PULL"
        path (str, optional): hint for the source location in storage or displaying.
        reshape (bool, optional): if True reshapes receiving array into 2d arrays/
        """
        Source.__init__(self, url=url, path=path, **kwargs)
        self.context = 0
        self.mode = mode
        self.ctx = None
        self.receiver = None
        self.pid = -1
        self.reshape = str_to_bool(str(reshape))

    def run(self, query):
        channels = query.get("channels", None)
        channel = channels[0] if (channels and len(channels)>0) else None
        try:
            self.connect()
            while not self.range.has_ended() and not self.aborted:
                data = self.receive()
                if not data:
                    raise Exception("Received None message.")
                if self.range.has_started():
                    pulse_id, array = data
                    name = channel if channel else (self.source if self.source else "Array10")
                    metadata = {} if self.reshape else {"width": self.shape[1], "height": self.shape[1]}
                    self.receive_channel(name, array, None, pulse_id if GENERATE_ID else None, check_changes=True, metadata=metadata)
        finally:
            self.close_channels()
            self.disconnect()

    def close(self):
        self.disconnect()
        Source.close(self)

    def connect(self):
        mode = zmq.PULL if self.mode == "PULL" else zmq.SUB
        self.ctx = zmq.Context()
        self.receiver = self.ctx.socket(mode)
        self.receiver.connect(self.url)
        if mode == zmq.SUB:
            self.receiver.subscribe("")
        self.message_count = 0

    def disconnect(self):
        try:
            self.receiver.close()
        except:
            pass
        try:
            self.ctx.term()
        except:
            pass
        finally:
            self.ctx = None

    def receive(self):
        try:
            header = self.receiver.recv()
            header = json.loads(''.join(chr(i) for i in header))
            self.shape = header.get("shape")
            self.dtype = header.get("type", "int8")
            self.source = header.get("source", "")
            data = self.receiver.recv()
            if data is not None:
                array = numpy.frombuffer(data, dtype=self.dtype)
                if self.reshape:
                    array = array.reshape(self.shape)
                self.pid = self.pid + 1
                return self.pid, array
        except Exception as e:
            _logger.warning("Error processing Array10: %s" % (str(e),))
            raise



import logging
import time
import numpy as np
from datahub import Consumer
from datahub.utils.timing import string_to_timestamp
from datahub import str_to_bool
from datahub.utils.plot import PlotClient

_logger = logging.getLogger(__name__)




class PShell(Consumer):
    def __init__(self,  channels=None, address="localhost", port=7777, timeout=3.0, layout="vertical", context=None,
                 style=None, colormap="viridis", color=None, marker_size=3, line_width=None, max_count=None, max_rate=None, **kwargs):
        Consumer.__init__(self, **kwargs)
        self.clients = {}
        self.plots = {}
        self.address = address
        self.port = port
        self.timeout = timeout
        self.channels = channels
        self.context = context
        self.style = style
        self.layout = layout.lower().capitalize() if layout else None
        self.colormap = colormap.lower().capitalize() if colormap else None
        self.color = color
        self.marker_size = marker_size
        self.line_width = line_width
        self.max_count = max_count
        self.min_interval = (1.0/max_rate) if max_rate else None
        self.last_plotted_record={}

        ps = PlotClient(address=self.address, port=self.port,  timeout=self.timeout)
        try:
            ps.get_contexts()
        except:
            raise Exception(f"Cannot connect to plot server on {address}:{port}")
        finally:
            ps.close()

    def on_close(self):
        for name, client in self.clients.items():
            client.close()

    def on_start(self, source):
        source_context = str_to_bool(str(self.context))==True
        pc = PlotClient(address=self.address, port=self.port, context=source.get_id() if source_context else self.context, timeout=self.timeout)
        self.clients[source] = pc
        pc.clear_plots()
        pc.set_context_attrs(quality=None, layout=self.layout)
        pc.set_progress(None)
        pc.set_status("Idle")

    def on_stop(self, source, exception):
        pass

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        if self.channels:
            if not name in self.channels:
                return
        pc = self.clients[source]
        if len(shape)==0:
            plot = pc.add_line_plot(name)
            pc.clear_plot(plot)
            series = pc.add_line_series(plot, name)
            pc.set_line_series_attrs(series, color=self.color, marker_size=self.marker_size, line_width=self.line_width, max_count=self.max_count)
            xdata = None
        elif len(shape) == 1:
            plot = pc.add_line_plot(name)
            pc.clear_plot(plot)
            series = pc.add_line_series(plot, name)
            pc.set_line_series_attrs(series, color=self.color, marker_size=2, line_width=1)
            xdata = list(range(shape[0]))
        elif len(shape) == 2:
            plot = pc.add_matrix_plot(name, style=self.style, colormap=self.colormap)
            series = pc.add_matrix_series(plot, "Matrix Series 1")
            pc.set_matrix_series_attrs(series, None, None, None, None, None, None)
            xdata = None
        else:
            _logger.warning("Unsupported shape for channel: " + name)
            return
        pc.clear_series(series)
        self.plots[name] = (plot, series, shape, xdata)
        pc.set_status("Running")
        pc.set_progress(0.5)

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        if name in self.plots:
            plot, series, shape, xdata = self.plots[name]
            pc = self.clients[source]
            try:
                if self.min_interval:
                    if len(shape) > 0: #Only downsample arrays
                        timespan = time.time() - self.last_plotted_record.get(series, 0.0)
                        if timespan < self.min_interval:
                            return
                if len(shape) == 0:
                    if type(timestamp) == str:
                        timestamp = string_to_timestamp(timestamp)
                    if isinstance(value, np.floating):  # Different scalar float types don't change header
                        value = float(value)
                    if isinstance(value, np.integer):  # Different scalar int types don't change header
                        value = int(value)
                    pc.append_line_series_data(series, timestamp, value, None)
                elif len(shape) == 1:
                    pc.set_line_series_data(series, xdata, value.tolist(),  None)
                elif len(shape) == 2:
                    pc.set_matrix_series_data(series, value.tolist(), None, None)
            except Exception as e:
                print("Error in plotting: " + str(e))
            if self.min_interval:
                self.last_plotted_record[series] = time.time()

    def on_channel_completed(self, source, name):
        pc = self.clients.get(source, None)
        if pc:
            if not pc.context.closed:
                pc.set_progress(None)
                pc.set_status("Done")







"""CLI interface """
import sys
import traceback

from datahub import *
import argparse
import logging
import json
import inspect
from datahub.utils.reflection import get_meta

logger = logging.getLogger(__name__)

SOURCE_SEPARATOR = "/"
CHANNEL_SEPARATOR = ","
EMPTY_SOURCE = [{}]

def run_json(task):
    try:
        if type(task) == str:
            task = json.loads(task)
        hdf5 = task.get("hdf5", None)
        if type(hdf5) == list:
            hdf5 = hdf5[0]
        txt = task.get("txt", None)
        if type(txt) == list:
            txt = txt[0]
        prnt = task.get("print", None)
        plot = task.get("plot", None)
        pshell = task.get("pshell", None)
        start = task.get("start", None)
        end = task.get("end", None)
        path = task.get("path", None)
        decompress = task.get("decompress", False)
        compression = task.get("compression", Compression.GZIP)
        parallel = task.get("parallel", None)
        interval = task.get("interval", None)
        bins = task.get("bins", None)
        last = task.get("last", None)
        modulo = task.get("modulo", None)
        filter = task.get("filter", None)
        search = task.get("search", None)
        verbose = task.get("verbose", None)
        prefix = task.get("prefix", None)
        query_id = task.get("id", False)
        query_time = task.get("time", False)
        time_type = task.get("timestamp", "nano")
        channels = task.get("channels", None)
        backend = task.get("backend", None)
        url = task.get("url", None)
        align = task.get("align", None)
        if compression == "lz4":
            compression = Compression.BITSHUFFLE_LZ4
        elif compression.lower() in ["null", "none"]:
            compression = None

        valid_sources = {}
        for name in KNOWN_SOURCES.keys():
            _name = name
            exec(f"{name} = task.get('{name}', None)")
            exec(f"no = 0 if {name} is None else len({name})")
            exec(f"for i in range(no): \n   if {name}[i] is not None: valid_sources['{name}_'+str(i)] = ({name}[i], KNOWN_SOURCES['{name}'])")

        consumers = []
        if hdf5 is not None:
            consumers.append(HDF5Writer(hdf5, default_compression=compression))
        if txt is not None:
            consumers.append(TextWriter(txt))
        if prnt is not None:
            consumers.append(Stdout())
        try:
            if pshell is not None:
                if pshell==True:
                    pshell={}
                consumers.append(PShell(**pshell))
        except Exception as ex:
            logger.exception(ex)
        try:
            if plot is not None:
                if plot==True:
                    plot={}
                consumers.append(Plot(**plot))
        except Exception as ex:
            logger.exception(ex)
        sources = []

        #If does nt have query arg, construct based on channels arg and start/end
        def get_query(source):
            nonlocal start, end, interval, modulo, prefix, channels, bins, last
            query = source.get("query", None)
            if query is None:
                source_channels = source.pop("channels", None)
                if source_channels is None:
                    source_channels =  [] if channels is None else channels
                if type(source_channels) == str:
                    source_channels = source_channels.split(CHANNEL_SEPARATOR)
                    source_channels = [s.lstrip("'\"").rstrip("'\"") for s in source_channels]
                query = {"channels": source_channels}
                query.update(source)
            if "start" not in query:
                query["start"] = start
            if "end" not in query:
                query["end"] = end
            if "interval" not in query:
                if interval:
                    query["interval"] = interval
            if "modulo" not in query:
                if modulo:
                    query["modulo"] = modulo
            if "prefix" not in query:
                if prefix:
                    query["prefix"] = prefix
            if "filter" not in query:
                if filter:
                    query["filter"] = filter
            if "bins" not in query:
                if bins:
                    query["bins"] = bins
            if "last" not in query:
                if last:
                    query["last"] = last

            force_id = False
            query_by_id = query_id
            if "id" in query:
                force_id = query_by_id = str_to_bool(str(query["id"]))

            query_by_time = False if force_id else query_time
            if "time" in query:
                query_by_time = str_to_bool(str(query["time"]))
                if query_by_time:
                    query_by_id = False

            for arg in "start", "end":
                    try:
                        if type(query[arg]) != str or not is_null_str(query[arg]):
                            if query_by_id:
                                query[arg] = int(query[arg])
                            elif query_by_time:
                                query[arg] = float(query[arg])
                    except:
                        pass
            if (query["start"] is None) and (query["end"] is None):
                return None

            try:
                if "bins" in query.keys():
                    query["bins"] = int(query["bins"])
            except Exception as ex:
                del query["bins"]

            try:
                if "last" in query.keys():
                    query["last"] = bool(query["last"])
            except Exception as ex:
                del query["last"]



            return query

        def add_source(cfg, src):
            nonlocal channels
            src.query = get_query(cfg)
            sources.append(src)

        #Create source removing constructor parameters from the query dictionary
        def get_source_constructor(cls, typ):
            nonlocal backend, url
            signature = inspect.signature(cls)
            pars = signature.parameters
            ret = cls.__name__+"("
            index = 0
            for name, par in pars.items():
                if par.kind != inspect.Parameter.VAR_KEYWORD:
                    if index > 0:
                        ret = ret + ", "
                    if par.default == inspect.Parameter.empty:
                        ret = ret + name + "=" + typ + ".pop('" + name + "')"
                    else:
                        default_val = par.default
                        if (name == "backend") and backend:
                            default_val = backend
                        if (name == "url") and url:
                            default_val = url
                        if type (default_val) == str:
                            dflt = "'" + default_val + "'"
                        else:
                            dflt = str(default_val)
                        ret = ret + name + "=" + typ + ".pop('" + name + "', " + dflt + ")"
                    index = index + 1
            if index > 0:
                ret = ret + ", "
            ret = ret + f"auto_decompress={str(decompress)}, time_type='{str(time_type)}', prefix='{str(prefix)}'"
            ret = ret + ")"
            return ret

        if len(valid_sources)==0:
            if channels or (search!=None):
                # Add default source
                valid_sources[DEFAULT_SOURCE+ "_0"] = ({},KNOWN_SOURCES[DEFAULT_SOURCE])

        for name, (cfg,source) in valid_sources.items():
            exec(f"{name} = cfg")
            constructor = eval("get_source_constructor(" + source.__name__ + ", '" + name + "')")
            exec('if ' + name + ' is not None: add_source(' + name + ', ' + constructor + ')')
        for source in sources:
            if verbose is not None:
                source.verbose = verbose
            if parallel is not None:
                source.parallel = parallel
            if path is not None:
                if source.path is None:
                    source.path = path

        if search is not None:
            if search == []:
                search = [""]
            for source in sources:
                # By default seach all available backends, unless requested specifically in command line
                source.set_backend(backend)
                try:
                    for regex in search:
                        source.print_search(regex)
                except:
                    logger.exception(f"Error searching source: {str(source)}")
        else:
            if align:
                merger = Merger(filter = filter)
                src = merger.to_source()
                for source in sources:
                    source.add_listener(merger)
                for consumer in consumers:
                    src.add_listener(consumer)
            else:
                for source in sources:
                    for consumer in consumers:
                        source.add_listener(consumer)

            for source in sources:
                if source is not None:
                    if source.query is None:
                        source.print_help()
                    else:
                        source.request(source.query, background=True)

            for source in sources:
                source.join()

    finally:
        cleanup()


def parse_args():
    """Parse cli arguments with argparse"""

    class CustomHelpFormatter(argparse.RawDescriptionHelpFormatter):
        def _format_action_invocation(self, action):
            if not action.option_strings:
                metavar, = self._metavar_formatter(action, action.dest)(1)
                return metavar
            else:
                parts = []
                    # if the Optional doesn't take a value, format is:
                #    -s, --long
                if action.nargs == 0:
                    parts.extend(action.option_strings)

                # if the Optional takes a value, format is:
                #    -s ARGS, --long ARGS
                # change to
                #    -s, --long ARGS
                else:
                    default = action.dest.upper()
                    args_string = self._format_args(action, default)
                    for option_string in action.option_strings:
                        # parts.append('%s %s' % (option_string, args_string))
                        parts.append('%s' % option_string)
                    parts[-1] += ' %s' % args_string
                return ', '.join(parts)

        def _format_args(self, action, default_metavar):
            if action.nargs == argparse.ZERO_OR_MORE:
                if action.metavar:
                    return '[{}]'.format(action.metavar)
                else:
                    return ""
            else:
                return action.metavar or default_metavar

    usage = "datahub [--GLOBAL_ARG_1 VALUE]...[--GLOBAL_ARG_N VALUE] [--<SOURCE 1>] [SOURCE_1_ARG_1 VALUE]...[SOURCE_1_ARG_N VALUE]...[--<SOURCE M>] [SOURCE_M_ARG_1 VALUE]...[SOURCE_M_ARG_N VALUE]"
    desc='Command line interface for DataHub  ' + datahub.version()
    epilog =f"Sources: {','.join(KNOWN_SOURCES.keys())}"
    epilog = epilog + f"\nConsumers: {','.join(KNOWN_CONSUMERS.keys())}"
    if DEFAULT_SOURCE:
        epilog = epilog + f"\nDefault Source (can be set by the env var DEFAULT_DATA_SOURCE): {DEFAULT_SOURCE}"
    epilog = epilog + f"\nThe source argument can be omited if only the default source is used."
    epilog = epilog + '\n\nFor source specific documentation:  datahub --<SOURCE>'
    parser = argparse.ArgumentParser(usage=usage, description=desc, prefix_chars='--', formatter_class=CustomHelpFormatter, epilog=epilog)
    parser.add_argument("-j", "--json", help="Complete query defined as JSON", required=False)

    for name, (abbr, cls) in KNOWN_CONSUMERS.items():
        meta = eval("get_meta(" + cls.__name__ + ")")
        eval(f'parser.add_argument("-{abbr}", "--{name}", metavar="{meta}", help="{name} options", required=False, nargs="*")')

    parser.add_argument("-v", "--verbose", action='store_true', help="Displays complete search results, not just channels names", required=False)
    parser.add_argument("-s", "--start", help="Relative or absolute start time or ID", required=False)
    parser.add_argument("-e", "--end", help="Relative or absolute end time or ID", required=False)
    parser.add_argument("-i", "--id", action='store_true', help="Force query by id", required=False)
    parser.add_argument("-t", "--time", action='store_true', help="Force query by time", required=False)
    parser.add_argument("-c", "--channels", help="Channel list (comma-separated)", required=False)
    parser.add_argument("-n", "--bins", help="Number of data bins", required=False)
    parser.add_argument("-u", "--url", help="URL of default source", required=False)
    parser.add_argument("-b", "--backend", help="Backend of default source (use \"null\" for all backends)", required=False)
    parser.add_argument("-a", "--align", action='store_true', help="Merge sources aligning the message ids",required=False)
    parser.add_argument("-r", "--filter", help="Sets a filter for data", required=False)
    parser.add_argument("-l", "--last",  action='store_true', help="Include last value before range", required=False)
    parser.add_argument("-tt", "--timestamp", help="Timestamp type: nano/int (default), sec/float or str", required=False)
    parser.add_argument("-cp", "--compression", help="Compression: gzip (default), szip, lzf, lz4 or none", required=False)
    parser.add_argument("-dc", "--decompress", action='store_true', help="Auto-decompress compressed images", required=False)
    parser.add_argument("-pl", "--parallel", action='store_true', help="Parallelize query if possible",required=False)
    parser.add_argument("-px", "--prefix", action='store_true', help="Add source ID to channel names", required=False)
    parser.add_argument("-pt", "--path", help="Path to data in the file", required=False)
    parser.add_argument("-sr", "--search", help="Search channel names given a pattern (instead of fetching data)", required=False , nargs="*")
    parser.add_argument("-di", "--interval", help="Downsampling interval between samples in seconds", required=False)
    parser.add_argument("-dm", "--modulo", help="Downsampling modulo of the samples", required=False)

    for name, source in KNOWN_SOURCES.items():
        meta = eval("get_meta(" + source.__name__ + ")")
        meta = f"channels {meta}start=None end=None"
        eval(f'parser.add_argument("--{name}", metavar="{meta}", help="{name} query arguments", required=False, nargs="*")')
    args = parser.parse_args()
    return parser, args

def get_full_argument_name(parser, abbr):
    for action in parser._actions:
        if '-' + abbr in action.option_strings:
            return action.dest
    return None

def _split_list(list, separator):
    result = []
    sublist = []
    for item in list:
        if item == separator:
            if sublist:
                result.append(sublist)
                sublist = []
        else:
            sublist.append(item)
    if sublist:
        result.append(sublist)
    return result

def print_help():
    print(f"DataHub {datahub.version()}")
    if DEFAULT_SOURCE:
        print("Default Source (can be set by the env var DEFAULT_DATA_SOURCE):")
        print(f"\t{DEFAULT_SOURCE}")
    print("Sources:")
    for source in KNOWN_SOURCES.keys():
        print(f"\t{source}")
    print("Consumers:")
    for consumer in KNOWN_CONSUMERS.keys():
        print(f"\t{consumer}")
    print(f"For help use the option:\n\t-h")
    print(f"For help on a specific source use the option:\n\t--<source_name>")


def main():
    """Main function"""
    if len(sys.argv) <= 1:
        print_help()
        return
    parser, args = parse_args()

    def parse_arg_dict(parser, val):
        ret = {}
        if val:
            for arg, val in zip(val[::2], val[1::2]):
                full_name = get_full_argument_name(parser, arg)
                if full_name:
                    arg = full_name
                try:
                    ret[arg] = json.loads(val)
                except:
                    ret[arg] = val
        return ret
    try:
        #if args.action == 'search':
        #    return search(args)
        if args.json:
            run_json(args.json)
        else:
            task={}
            task["hdf5"] = args.hdf5
            task["txt"] = args.txt
            task["print"] = None if args.print is None else bool(args.print)
            if args.plot is not None:
                task["plot"] = parse_arg_dict(parser, args.plot)
            if args.pshell is not None:
                task["pshell"] = parse_arg_dict(parser, args.pshell)
            if args.start:
                task["start"] = args.start
            if args.end:
                task["end"] = args.end
            if args.id:
                task["id"] = bool(args.id)
            if args.time:
                task["time"] = bool(args.time)
            if args.timestamp:
                task["timestamp"] = args.timestamp
            if args.path:
                task["path"] = args.path
            if args.decompress:
                task["decompress"] = bool(args.decompress)
            if args.compression:
                task["compression"] = args.compression
            if args.parallel:
                task["parallel"] = args.parallel
            if args.interval:
                task["interval"] = args.interval
            if args.bins:
                task["bins"] = args.bins
            if args.last:
                task["last"] = bool(args.last)
            if args.modulo:
                task["modulo"] = args.modulo
            if args.filter:
                task["filter"] = args.filter
            if args.search is not None:
                task["search"] = args.search
            if args.verbose is not None:
                task["verbose"] = args.verbose
            if args.align is not None:
                task["align"] = args.align
            if args.prefix is not None:
                task["prefix"] = args.prefix
            if args.channels is not None:
                task["channels"] = args.channels
            if args.backend is not None:
                task["backend"] = args.backend
            if args.url is not None:
                task["url"] = args.url

            for source in KNOWN_SOURCES.keys():
                source_str = eval("args." + source)
                if type(source_str) == list:
                    if len(source_str) == 1:
                        task[source] = json.loads(source_str[0])
                    else:
                        task[source] = []
                        sources = _split_list(source_str, SOURCE_SEPARATOR)
                        if sources == []:
                            task[source] = EMPTY_SOURCE #if source is entered with no arguments, print help for it
                        for src in sources:
                            task[source].append(parse_arg_dict(parser, src))
            run_json(task)

    except RuntimeError as e:
        logger.error(e)
    return 0
if __name__ == '__main__':
    """
    json_str = '{' \
               '"file": "/Users/gobbo_a/dev/back/json.h5", ' \
               '"print": true, ' \
               '"epics": {"url": null, "query":{"start":null, "end":3.0, "channels": ["TESTIOC:TESTCALCOUT:Input", "TESTIOC:TESTSINUS:SinCalc", "TESTIOC:TESTWF2:MyWF"]}},' \
               '"bsread": {"url": "tcp://localhost:9999", "mode":"PULL", "query":{"start":null, "end":3.0, "channels":  ["UInt8Scalar", "Float32Scalar"]}}' \
               '}'
    args = ["--json", json_str]
    sys.argv=sys.argv + args
    """
    """
    args = ["--search", "SARFE10-PSSS059:SPECTRUM_X", "--databuffer"]
    sys.argv = sys.argv + args
    """
    """
    args = ["-h"]
    sys.argv = sys.argv + args
    args = ["--daqbuf"]
    sys.argv = sys.argv + args
    """
    main()




import unittest
from datahub import *

backend = "sf-databuffer"


filename = "/Users/gobbo_a/dev/back/retrieval.h5"
#channels = ["SARFE10-PSSS059:FIT-COM", "SARFE10-PSSS059:FIT-FWHM"]
channels = ["S10BC01-DBPM010:Q1"]
start = "2024-02-14 08:50:00.000"
end = "2024-02-14 08:50:05.000"

#String channels not working
#channels = ["SARES20-CAMS142-M5.processing_parameters"]
#start = "2024-02-14 16:50:00.000"
#end = "2024-02-14 16:50:05.000"

#start = '2024-02-12T07:50:00.000000Z'
#end =  '2024-02-12T07:50:01.000000Z'

"""
query = {
    "channels": channels,
    "range": {
        "type": "date",
        "startDate": start,
        "endDate": end
    }
}
"""

query = {
    "channels": channels,
    "start": start,
    "end": end
}





class RetrivalTest(unittest.TestCase):
    def setUp(self):
        self.source = Retrieval(backend=backend)

    def tearDown(self):
        self.source.close()


    def test_listeners(self):
        hdf5 = HDF5Writer(filename)
        stdout = Stdout()
        table = Table()
        self.source.add_listener(hdf5)
        self.source.add_listener(stdout)
        self.source.add_listener(table)
        self.source.request(query)
        dataframe = table.as_dataframe()
        self.assertEqual(list(dataframe.keys()), channels)
        self.source.close_listeners()


if __name__ == '__main__':
    unittest.main()
